"""Gene related processes"""
from pyppl import Proc, Box
from . import params
from .seq import _pPromoters
#from .utils import genenorm, write
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pGenePromoters():
	"""
	@name:
		pGenePromoters
	@description:
		Alias of `seq.pPromoters`.
	"""
	pGenePromoters = _pPromoters()
	return pGenePromoters

@procfactory
def _pGeneNameNorm():
	"""
	@name:
		pGeneNameNorm
	@description:
		Normalize gene names using MyGeneinfo.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`inopts`: options for reading input file.
		`outopts`: options for writing output file.
			- `query` : Output the original query column? Default: `False`
			- `cnames`: Output headers? Default: `True`
		`notfound`: What if a symbol is not found. Default: ignore
			- skip  : skip the record(don't write it to output file)
			- ignore: use the original name;
			- error : report error
		`genecol` : the column index containing the gene names
		`frm`     : the original format. Default: 'symbol, alias'
		`to`      : the output gene name format. Default: 'symbol'
		`genome`  : the genome. Default: 'hg19'
		`cachedir`: The cache directory
	"""
	pGeneNameNorm               = Proc(desc = 'Normalize gene names using MyGeneinfo.')
	pGeneNameNorm.input         = 'infile:file'
	pGeneNameNorm.output        = 'outfile:file:{{i.infile | bn}}'
	pGeneNameNorm.errhow        = 'retry'
	pGeneNameNorm.args.notfound = 'ignore'
	pGeneNameNorm.args.inopts   = Box(skip = 0, comment = '#', delimit = '\t')
	pGeneNameNorm.args.outopts  = Box(delimit = '\t', cnames = True, query = False)
	pGeneNameNorm.args.genecol  = ''
	pGeneNameNorm.args.frm      = 'symbol, alias'
	pGeneNameNorm.args.to       = 'symbol'
	pGeneNameNorm.args.genome   = params.genome.value
	pGeneNameNorm.args.cachedir = params.cachedir.value
	pGeneNameNorm.lang          = params.python.value
	pGeneNameNorm.script        = "file:scripts/gene/pGeneNameNorm.py"
	return pGeneNameNorm

@procfactory
def _pIPI():
	"""
	@name:
		pIPI
	@description:
		Convert gene symbol to IPI protein accession and vice versa.
		One gene symbol could map to multiple IPIs, which will be separated by pipe (|)
	@input:
		`infile:file` : The input file
	@output:
		`outfile:file`: The output file
	@args:
		`notfound`: What if a record is not found: Default: `ignore`
			- `skip`  : skip the record(don't write it to output file)
			- `ignore`: use the original name;
			- `error` : report error
		`genecol`: The column index containing the gene/protein record
		`ipidb`: The IPI xref database (see http://ftp.ebi.ac.uk/pub/databases/IPI/last_release/current/).
		`fromipi`: Whether the input is IPI or genes
		`inopts`: The options for input file
		`outopts`: The options for output file
	"""
	pIPI               = Proc(desc = 'Convert gene symbol to IPI protein accession and vice versa.')
	pIPI.input         = 'infile:file'
	pIPI.output        = 'outfile:file:{{i.infile | bn}}'
	pIPI.errhow        = 'retry'
	pIPI.args.notfound = 'ignore'
	pIPI.args.inopts   = Box(skip = 0, comment = '#', delimit = '\t')
	pIPI.args.outopts  = Box(delimit = '\t', headDelimit = '\t', headPrefix = '', headTransform = None, head = True, query = False)
	pIPI.args.genecol  = None
	pIPI.args.fromipi  = True
	pIPI.args.ipidb    = params.ipidb.value
	pIPI.lang          = params.python.value
	pIPI.script        = "file:scripts/gene/pIPI.py"
	return pIPI

@procfactory
def _pGeneTss():
	"""
	@name:
		pGeneTss
	@description:
		Get gene TSS in BED format.
	@input:
		`infile:file`: The input file containing genes
	@output:
		`outfile:file`: The output BED file
	@args:
		`notfound`: What if the gene is not found. Default: skip.
			- error: report error
		`header`: Whether the input file contains header. Default: False
		`skip`: Skip N lines of input file. Default: 0
			- This has highest priority of header and comment
		`comment`: The comment line start sign. Default: #
		`delimit`: The delimit of input file if it has multiple column. Default: `\\t`
		`col`: The column index contains the genes. Default: 0
		`frm`: The format of the genes. Default: `symbol, alias`
		`tmpdir`: The tmpdir used to store mygene cache files.
		`genome`: In which genome to fetch the coordinates. Default: hg19
	"""
	pGeneTss                = Proc(desc = 'Get gene TSS in BED format')
	pGeneTss.input          = 'infile:file'
	pGeneTss.output         = 'outfile:file:{{i.infile | fn}}-tss.bedx'
	pGeneTss.errhow         = 'retry'
	pGeneTss.args.notfound  = 'skip' # error
	pGeneTss.args.genecol   = ''
	pGeneTss.args.inopts    = Box(skip = 0, comment = '#', delimit = '\t')
	pGeneTss.args.outopts   = Box(delimit = '\t', headDelimit = '\t', headPrefix = '', headTransform = None, head = False, query = False, ftype = 'bed')
	pGeneTss.args.frm       = 'symbol, alias'
	pGeneTss.args.cachedir  = params.cachedir.value
	pGeneTss.args.genome    = params.genome.value
	#pGeneTss.envs.genenorm  = genenorm.py
	#pGeneTss.envs.writeBedx = write.bedx.py
	pGeneTss.lang           = params.python.value
	pGeneTss.script         = "file:scripts/gene/pGeneTss.py"
	return pGeneTss

@procfactory
def _pGeneBody():
	"""
	@name:
		pGeneBody
	@description:
		Get gene body region in BED format
	@input:
		`infile:file`: The input file containing genes
	@output:
		`outfile:file`: The gene body region
	@args:
		`notfound`: What if a gene is not found when transfer the gene names to gene symbols
			- error: report error
			- skip (default): skip it
		`inmeta`:   The metadata for input file, mainly to indicate where the GENE column is.
		`inopts`:   Input options for reading input file.
			- skip: number of lines to skip. Default: 0
			- comment: the starting string for comment lines. Default: #
			- delimit: The delimit for the input file. Default: '\\t'
		frm: The gene name format in the input file. Default: 'symbol, alias'
		tmpdir: The tmpdir to cache the gene name conversion.
		genome: The genome used to do the conversion.
	"""
	pGeneBody               = Proc(desc = 'Get gene body in BED format')
	pGeneBody.input         = 'infile:file'
	pGeneBody.output        = 'outfile:file:{{i.infile | fn}}-body.bed'
	pGeneBody.args.inopts   = Box(cnames = False)
	pGeneBody.args.notfound = 'skip' # error
	pGeneBody.args.genecol  = ''
	pGeneBody.args.refgene  = params.refgene.value
	pGeneBody.lang          = params.python.value
	pGeneBody.script        = "file:scripts/gene/pGeneBody.py"
	return pGeneBody




from os import path

infile = {{i.infile | quote}}
outdir = {{o.outdir | quote}}
n      = {{args.n}}
header = {{args.cnames}}
nskip  = {{args.skip}}

# get # lines and header
nlines   = 0
headerow = ''
with open(infile) as f:
	for line in f: 
		if header and nlines == nskip: headerow = line
		nlines += 1
		
nlines -= nskip
if header: nlines -= 1

nrow, remain = divmod(nlines, n)
nrows = [0] * n
for i in range(n):
	nrows[i] = nrow + 1 if i < remain else nrow

with open(infile) as f:
	for _ in range(nskip): f.readline()
	if header: f.readline()
	for i in range(n):
		subfile = path.join(outdir, "{{i.infile | fn}}-%s{{i.infile | ext}}" % (i+1))
		with open(subfile, 'w') as fout:
			if header: fout.write(headerow)
			for _ in range(nrows[i]):
				fout.write(f.readline())




from pyppl import Box
from enrichr import Enrichr
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

{%python from bioprocs.utils import alwaysList%}
infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
inopts  = {{args.inopts | repr}}
genecol = {{args.genecol | repr}}
libs    = {{args.libs | alwaysList}}

reader   = TsvReader(infile, **inopts)
genes    = reader.dump(genecol)
genes    = list(set(genes))
enrich   = Enrichr()
pathways = {}

writer = TsvWriter(outfile)
if inopts.cnames:
	writer.cnames = reader.cnames + ['Pathway_' + lib for lib in libs]
	writer.writeHead()

reader.rewind()
for r in reader:
	gene = r[genecol]
	if gene not in pathways:
		pathways[gene] = {lib:enlib.terms
			for lib, enlib in enrich.genemap(gene, libs).items()}

	for lib in libs:
		r['Pathway_' + lib] = ' | '.join(pathways[gene][lib]) \
			if lib in pathways[gene] else ''
	writer.write(r.values())

reader.close()
writer.close()



from pyppl import ProcSet
from bioprocs.common import pStr2File, pFile2Proc, pSort
from bioprocs.tsv import pTsvJoin
from bioprocs.seq import pPromoters, pConsv, pConsvPerm
from bioprocs.bed import pBedGetfasta
from bioprocs.tfbs import pMotifScan
from bioprocs import params

"""
@name:
	aTfbsTfP
@description:
	Scan motifs on genes' promoter regions by giving TF names.
@depends:
		pPromoters[*] \
					|
				pBedGetfasta \
								pMotifScan[!]
	     pTFs[*] -- pTsvJoin /
						|
		    pTFList[*] /
@input:
	- TF list file, one per line
	- gene list file, one per line
	- TF (1st col) list file with motif ids (2nd col). Default: params.tflist.value
"""
aTfbsTfP = ProcSet(
	pSort.copy('pTFs'),
	pPromoters,
	pSort.copy('pTFList'),
	pTsvJoin,
	pBedGetfasta,
	pMotifScan,
	depends = False
)
# defaults
aTfbsTfP.pPromoters.runner = 'local'
#delegate
aTfbsTfP.delegate('args.up'      , 'pPromoters')
aTfbsTfP.delegate('args.down'    , 'pPromoters')
aTfbsTfP.delegate('args.genome'  , 'pPromoters')
aTfbsTfP.delegate('args.ref'     , 'pBedGetfasta')
aTfbsTfP.delegate('args.pval'    , 'pMotifScan')
aTfbsTfP.delegate('args.tfmotifs', 'pMotifScan')
# depends
aTfbsTfP.starts               = aTfbsTfP.pTFs, aTfbsTfP.pPromoters, aTfbsTfP.pTFList
aTfbsTfP.ends                 = aTfbsTfP.pMotifScan
aTfbsTfP.pMotifScan.depends   = aTfbsTfP.pTsvJoin, aTfbsTfP.pBedGetfasta
aTfbsTfP.pBedGetfasta.depends = aTfbsTfP.pPromoters
aTfbsTfP.pTsvJoin.depends     = aTfbsTfP.pTFList, aTfbsTfP.pTFs
# input
aTfbsTfP.pTFList.input    = [params.tflist.value]
# input size of either pTFs or pTFList should be 1
aTfbsTfP.pTsvJoin.input   = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)).flatten() for l in [max(ch1.length(), ch2.length())]]
aTfbsTfP.pMotifScan.input = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)) for l in [max(ch1.length(), ch2.length())]][0]
# args
aTfbsTfP.pBedGetfasta.args.params.name = True
aTfbsTfP.pTFList.args.params.k         = 2
aTfbsTfP.pTsvJoin.args.match           = 'lambda line1, line2: -1 if line1[1] == line2[0] else 0 if line1[1] < line2[0] else 1'
aTfbsTfP.pTsvJoin.args.do              = 'lambda line1, line2: fout.write("\\t".join(line1) + "\\n")'

"""
@name:
	aTfbsP
@description:
	Scan motifs on genes' promoter regions.
@depends:
	pPromoters[*] \
					pBedGetfasta \
								  pMotifScan[!]
					  pTFList[*] /
@input:
	- TF (1st col) list file with motif ids (2nd col). Default: params.tflist.value
	- gene list file, one per line
"""
aTfbsP = ProcSet(
	pFile2Proc.copy('pTFList'),
	pPromoters,
	pBedGetfasta,
	pMotifScan,
	depends = False
)
# defaults
aTfbsP.pTFList.runner    = 'local'
aTfbsP.pPromoters.runner = 'local'
# delegate
aTfbsP.delegate('args.up', 'pPromoters')
aTfbsP.delegate('args.down', 'pPromoters')
aTfbsP.delegate('args.genome', 'pPromoters')
aTfbsP.delegate('args.ref', 'pBedGetfasta')
aTfbsP.delegate('args.pval', 'pMotifScan')
aTfbsP.delegate('args.tfmotifs', 'pMotifScan')
# depends
aTfbsP.starts               = aTfbsP.pTFList,     aTfbsP.pPromoters
aTfbsP.ends                 = aTfbsP.pMotifScan
aTfbsP.pMotifScan.depends   = aTfbsP.pTFList,     aTfbsP.pBedGetfasta
aTfbsP.pBedGetfasta.depends = aTfbsP.pPromoters
# args
aTfbsP.pBedGetfasta.args.params.name = True


"""
@name:
	aTfbsTfR
@description:
	Scan motifs on a given regions.
@depends:
	                      pBedGetfasta[*]  \
						                     pMotifScan[!]
		   	           pTFs[*] -- pTsvJoin /
	                          |
	          pSortTFList[*] /
@input:
	- TF list file, one per line
	- region file in bed
	- TF (1st col) list file with motif ids (2nd col). Default: params.tflist.value
"""
aTfbsTfR = ProcSet(
	pSort.copy('pTFs'),
	pBedGetfasta,
	pSort.copy('pTFList'),
	pTsvJoin,
	pMotifScan,
	depends = False
)
# defaults
aTfbsTfR.pTFList.runner     = 'local'
# delegate
aTfbsTfR.delegate('args.ref', 'pBedGetfasta')
aTfbsTfR.delegate('args.pval', 'pMotifScan')
aTfbsTfR.delegate('args.tfmotifs', 'pMotifScan')
# depends
aTfbsTfR.starts              = aTfbsTfR.pTFs, aTfbsTfR.pBedGetfasta, aTfbsTfR.pTFList
aTfbsTfR.ends                = aTfbsTfR.pMotifScan
aTfbsTfR.pMotifScan.depends  = aTfbsTfR.pTsvJoin, aTfbsTfR.pBedGetfasta
aTfbsTfR.pTsvJoin.depends    = aTfbsTfR.pTFList, aTfbsTfR.pTFs
# input
aTfbsTfR.pTFList.input    = [params.tflist.value]
aTfbsTfR.pTsvJoin.input   = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)).flatten() for l in [max(ch1.length(), ch2.length())]]
aTfbsTfR.pMotifScan.input = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)) for l in [max(ch1.length(), ch2.length())]][0]
# args
aTfbsTfR.pBedGetfasta.args.params.name = True
aTfbsTfR.pTFList.args.params.k         = 2
aTfbsTfR.pTsvJoin.args.match           = 'lambda line1, line2: -1 if line1[1] == line2[0] else 0 if line1[1] < line2[0] else 1'
aTfbsTfR.pTsvJoin.args.do              = 'lambda line1, line2: fout.write("\\t".join(line1) + "\\n")'


"""
@name:
	aTfbsR
@description:
	Scan motifs on a given regions.
@depends:
	pBedGetfasta[*]  \
						pMotifScan[!]
		   pTFList[*] /
@input:
	- TF (1st col) list file with motif ids (2nd col). Default: params.tflist.value
	- region file in bed
"""
aTfbsR = ProcSet(
	pFile2Proc.copy('pTFList'),
	pBedGetfasta,
	pMotifScan,
	depends = False
)
# defaults
aTfbsR.pTFList.runner     = 'local'
# delegate
aTfbsR.delegate('args.ref', 'pBedGetfasta')
aTfbsR.delegate('args.pval', 'pMotifScan')
aTfbsR.delegate('args.tfmotifs', 'pMotifScan')
# depends
aTfbsR.starts             = aTfbsR.pTFList, aTfbsR.pBedGetfasta
aTfbsR.ends               = aTfbsR.pMotifScan
aTfbsR.pMotifScan.depends = aTfbsR.pTFList, aTfbsR.pBedGetfasta
# args
aTfbsR.pBedGetfasta.args.params.name = True

"""
@name:
	aTfbsPC
@description:
	Scan motifs on genes' promoter regions with conservation.
@depends:
	pPromoters[*] \
					pBedGetfasta \
								  pMotifScan
					   pTFList[*] /	         \
					                          pConsv[!]
											 /
						    	pConsvPerm[*]
@input:
	- TF list file, one per line
	- gene list file, one per line
	- Seeds for pConsvPerm. Default: [0]
"""
aTfbsPC = ProcSet(
	pFile2Proc.copy('pTFList'),
	pPromoters,
	pConsvPerm,
	pBedGetfasta,
	pMotifScan,
	pConsv,
	depends = False
)
# defaults
aTfbsPC.pTFList.runner    = 'local'
aTfbsPC.pPromoters.runner = 'local'
aTfbsPC.pConsvPerm.input  = [0]
# delegate
aTfbsPC.delegate('args.up', 'pPromoters')
aTfbsPC.delegate('args.down', 'pPromoters')
aTfbsPC.delegate('args.genome', 'pPromoters')
aTfbsPC.delegate('args.ref', 'pBedGetfasta')
aTfbsPC.delegate('args.pval', 'pMotifScan')
aTfbsPC.delegate('args.tfmotifs', 'pMotifScan')
# aTfbsPC.delegate('args.cpval', 'pConsv', 'args.pval')
aTfbsPC.delegate('args.len', 'pConsvPerm')
aTfbsPC.delegate('args.nperm', 'pConsvPerm')
aTfbsPC.delegate('args.chrsizes', 'pConsvPerm')
# depends
aTfbsPC.starts               = aTfbsPC.pTFList, aTfbsPC.pPromoters, aTfbsPC.pConsvPerm
aTfbsPC.ends                 = aTfbsPC.pConsv
aTfbsPC.pConsv.depends       = aTfbsPC.pMotifScan, aTfbsPC.pConsvPerm
aTfbsPC.pMotifScan.depends   = aTfbsPC.pTFList, aTfbsPC.pBedGetfasta
aTfbsPC.pBedGetfasta.depends = aTfbsPC.pPromoters
# input
aTfbsPC.pConsv.input = lambda ch1, ch2: ch1.outfile.cbind(ch2)
# args
aTfbsPC.pBedGetfasta.args.params.name = True
aTfbsPC.pConsv.args.pval              = 0.05


"""
@name:
	aTfbsRC
@description:
	Scan motifs on a given regions with conservation score.
@depends:
	pBedGetfasta[*]  \
						pMotifScan
		   pTFList[*] /             \
		                            pConsv[!]
				             	   /
					  pConsvPerm[*]
@input:
	- TF list file, one per line
	- region list in bed.
	- Seeds for pConsvPerm. Default: [0]
"""
aTfbsRC = ProcSet(
	pFile2Proc.copy('pTFList'),
	pBedGetfasta,
	pConsvPerm,
	pMotifScan,
	pConsv,
	depends = False
)
# defaults
aTfbsRC.pTFList.runner     = 'local'
aTfbsRC.pConsvPerm.input  = [0]
# delegate
aTfbsRC.delegate('args.ref', 'pBedGetfasta')
aTfbsRC.delegate('args.pval', 'pMotifScan')
aTfbsRC.delegate('args.tfmotifs', 'pMotifScan')
# aTfbsRC.delegate('args.cpval', 'pConsv', 'args.pval')
aTfbsRC.delegate('args.len', 'pConsvPerm')
aTfbsRC.delegate('args.nperm', 'pConsvPerm')
aTfbsRC.delegate('args.chrsizes', 'pConsvPerm')
# depends
aTfbsRC.starts             = aTfbsRC.pTFList, aTfbsRC.pBedGetfasta, aTfbsRC.pConsvPerm
aTfbsRC.ends               = aTfbsRC.pConsv
aTfbsRC.pConsv.depends     = aTfbsRC.pMotifScan, aTfbsRC.pConsvPerm
aTfbsRC.pMotifScan.depends = aTfbsRC.pTFList, aTfbsRC.pBedGetfasta
# input
aTfbsRC.pConsv.input = lambda ch1, ch2: ch1.outfile.cbind(ch2)
# args
aTfbsRC.pBedGetfasta.args.params.name = True
aTfbsRC.pConsv.args.pval              = 0.05

"""
@name:
	aTfbsTfPC
@description:
	Scan motifs on genes' promoter regions by giving TF names with conservation.
@depends:
		pPromoters[*] \
					|
				pBedGetfasta \
								pMotifScan
	    pTFs[*] -- pTsvJoin /               \
						|                    \
		    pTFList[*] /                       pConsv[!]
			                   pConsvPerm[*] /
@input:
	- TF list file, one per line
	- gene list file, one per line
	- TF (1st col) list file with motif ids (2nd col). Default: params.tflist.value
"""
aTfbsTfPC = ProcSet(
	pSort.copy('pTFs'),
	pPromoters,
	pSort.copy('pTFList'),
	pTsvJoin,
	pBedGetfasta,
	pConsvPerm,
	pMotifScan,
	pConsv,
	depends = False
)
# defaults
aTfbsTfPC.pPromoters.runner = 'local'
aTfbsTfPC.pConsvPerm.input  = [0]
# delegate
aTfbsTfPC.delegate('args.up'      , 'pPromoters')
aTfbsTfPC.delegate('args.down'    , 'pPromoters')
aTfbsTfPC.delegate('args.genome'  , 'pPromoters')
aTfbsTfPC.delegate('args.ref'     , 'pBedGetfasta')
aTfbsTfPC.delegate('args.pval'    , 'pMotifScan')
aTfbsTfPC.delegate('args.tfmotifs', 'pMotifScan')
# aTfbsTfPC.delegate('args.cpval'   , 'pConsv', 'args.pval')
aTfbsTfPC.delegate('args.len'     , 'pConsvPerm')
aTfbsTfPC.delegate('args.nperm'   , 'pConsvPerm')
aTfbsTfPC.delegate('args.chrsizes', 'pConsvPerm')
# depends
aTfbsTfPC.starts               = aTfbsTfPC.pTFs, aTfbsTfPC.pPromoters, aTfbsTfPC.pTFList, aTfbsTfPC.pConsvPerm
aTfbsTfPC.ends                 = aTfbsTfPC.pConsv
aTfbsTfPC.pConsv.depends       = aTfbsTfPC.pMotifScan, aTfbsTfPC.pConsvPerm
aTfbsTfPC.pMotifScan.depends   = aTfbsTfPC.pTsvJoin, aTfbsTfPC.pBedGetfasta
aTfbsTfPC.pBedGetfasta.depends = aTfbsTfPC.pPromoters
aTfbsTfPC.pTsvJoin.depends     = aTfbsTfPC.pTFList, aTfbsTfPC.pTFs
# input
aTfbsTfPC.pTFList.input    = [params.tflist.value]
# input size of either pTFs or pTFList should be 1
aTfbsTfPC.pTsvJoin.input   = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)).flatten() for l in [max(ch1.length(), ch2.length())]]
aTfbsTfPC.pMotifScan.input = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)) for l in [max(ch1.length(), ch2.length())]][0]
aTfbsTfPC.pConsv.input     = lambda ch1, ch2: ch1.outfile.cbind(ch2)
# args
aTfbsTfPC.pBedGetfasta.args.params.name = True
aTfbsTfPC.pTFList.args.params.k         = 2
aTfbsTfPC.pTsvJoin.args.match           = 'lambda line1, line2: -1 if line1[1] == line2[0] else 0 if line1[1] < line2[0] else 1'
aTfbsTfPC.pTsvJoin.args.do              = 'lambda line1, line2: fout.write("\\t".join(line1) + "\\n")'
aTfbsTfPC.pConsv.args.pval              = 0.05

"""
@name:
	aTfbsTfRC
@description:
	Scan motifs on a given regions with conservation.
@depends:
	                      pBedGetfasta[*]  \
						                     pMotifScan
		   	            pTFs[*] -- pTsvJoin /         \
	                          |                        \
	          pSortTFList[*] /                           pConsv[!]
			                            pConsvPerm[*]  /
@input:
	- TF list file, one per line
	- region file in bed
	- TF (1st col) list file with motif ids (2nd col). Default: params.tflist.value
"""
aTfbsTfRC = ProcSet(
	pSort.copy('pTFs'),
	pBedGetfasta,
	pConsvPerm,
	pSort.copy('pTFList'),
	pTsvJoin,
	pMotifScan,
	pConsv,
	depends = False
)
# defaults
aTfbsTfRC.pTFList.runner    = 'local'
aTfbsTfRC.pConsvPerm.input  = [0]
# delegate
aTfbsTfRC.delegate('args.ref', 'pBedGetfasta')
aTfbsTfRC.delegate('args.pval', 'pMotifScan')
aTfbsTfRC.delegate('args.tfmotifs', 'pMotifScan')
# aTfbsTfRC.delegate('args.cpval'   , 'pConsv', 'args.pval')
aTfbsTfRC.delegate('args.len'     , 'pConsvPerm')
aTfbsTfRC.delegate('args.nperm'   , 'pConsvPerm')
aTfbsTfRC.delegate('args.chrsizes', 'pConsvPerm')
# depends
aTfbsTfRC.starts             = aTfbsTfRC.pTFs,   aTfbsTfRC.pBedGetfasta, aTfbsTfRC.pTFList, aTfbsTfRC.pConsvPerm
aTfbsTfRC.ends               = aTfbsTfRC.pConsv
aTfbsTfRC.pConsv.depends     = aTfbsTfRC.pMotifScan, aTfbsTfRC.pConsvPerm
aTfbsTfRC.pMotifScan.depends = aTfbsTfRC.pTsvJoin,   aTfbsTfRC.pBedGetfasta
aTfbsTfRC.pTsvJoin.depends   = aTfbsTfRC.pTFList,    aTfbsTfRC.pTFs
# input
aTfbsTfRC.pTFList.input    = [params.tflist.value]
aTfbsTfRC.pTsvJoin.input   = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)).flatten() for l in [max(ch1.length(), ch2.length())]]
aTfbsTfRC.pMotifScan.input = lambda ch1, ch2: [ch1.repRow(l).cbind(ch2.repRow(l)) for l in [max(ch1.length(), ch2.length())]][0]
aTfbsTfRC.pConsv.input     = lambda ch1, ch2: ch1.outfile.cbind(ch2)
# args
aTfbsTfRC.pBedGetfasta.args.params.name = True
aTfbsTfRC.pTFList.args.params.k         = 2
aTfbsTfRC.pTsvJoin.args.match           = 'lambda line1, line2: -1 if line1[1] == line2[0] else 0 if line1[1] < line2[0] else 1'
aTfbsTfRC.pTsvJoin.args.do              = 'lambda line1, line2: fout.write("\\t".join(line1) + "\\n")'
aTfbsTfRC.pConsv.args.pval              = 0.05




# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['bioprocs',
 'bioprocs.console',
 'bioprocs.console.pipeline',
 'bioprocs.scripts',
 'bioprocs.sets',
 'bioprocs.utils']

package_data = \
{'': ['*'],
 'bioprocs': ['reports/bcftools/*',
              'reports/gsea/*',
              'reports/imtherapy/*',
              'reports/mlearn/*',
              'reports/plot/*',
              'reports/rnaseq/*',
              'reports/stats/*',
              'reports/tcgamaf/*',
              'reports/tumhet/*'],
 'bioprocs.scripts': ['algorithm/*',
                      'bcftools/*',
                      'bed/*',
                      'bedtools/*',
                      'bwtool/*',
                      'chipseq/*',
                      'cluster/*',
                      'cnvkit/*',
                      'common/*',
                      'docx/*',
                      'eqtl/*',
                      'fastx/*',
                      'gene/*',
                      'genomeplot/*',
                      'gsea/*',
                      'hic/*',
                      'imtherapy/*',
                      'marray/*',
                      'misc/*',
                      'mlearn/*',
                      'network/*',
                      'picard/*',
                      'plink/*',
                      'plot/*',
                      'power/*',
                      'rank/*',
                      'resource/*',
                      'rnaseq/*',
                      'sambam/*',
                      'seq/*',
                      'snp/*',
                      'snparray/*',
                      'sql/*',
                      'stats/*',
                      'tabix/*',
                      'tcga/*',
                      'tcgamaf/*',
                      'tfbs/*',
                      'tsv/*',
                      'tumhet/*',
                      'vcf/*',
                      'web/*',
                      'xlsx/*']}

install_requires = \
['PyPPL', 'completions', 'pyparam', 'python-box87']

entry_points = \
{'console_scripts': ['bioprocs = bioprocs.console:main']}

setup_kwargs = {
    'name': 'bioprocs',
    'version': '0.1.2',
    'description': 'A set of PyPPL processes for bioinformatics.',
    'long_description': None,
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
Some sql database utilities
"""

from pyppl import Proc

from . import params
from .utils import sql
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pCreateTable():
	"""
	@name:
		pCreateTable
	@description:
		Create tables in the database
	@input:
		`dsn`: The dsn to connect to the database
			- currently support `sqlite:file=...`
		`schema:file`: The schema file
			- could be a pure schema file:
			```
			Field	Type	Statement
			ID	INT	PRIMARY KEY
			...
			```
			- or a data file with header
	@output:
		`dsn`: The dsn
	@args:
		`intype`: The input file schema file or a data file. Default: `schema`
		`drop`:  Force creating the table (drop the pre-existing table)
		`delimit`:The delimit of input file. Default: `\\t`
	"""
	pCreateTable                  = Proc(desc = 'Create tables in the database.')
	pCreateTable.input            = "dsn, schema:file"
	pCreateTable.output           = "dsn:var:{{i.dsn}}"
	pCreateTable.args.intype      = 'schema'
	pCreateTable.args.drop        = False
	pCreateTable.args.delimit     = "\t"
	pCreateTable.envs.dsnparse    = sql.dsnparse
	pCreateTable.envs.schemaparse = sql.schemaparse
	pCreateTable.lang             = params.python.value
	pCreateTable.script           = 'file:scripts/sql/pCreateTable.py'
	return pCreateTable

@procfactory
def _pImportData():
	"""
	@name:
		pImportData
	@description:
		Create tables and import the data
	@input:
		`dsn`: The dsn to connect to the database
			- currently support `sqlite:file=...`
		`datafile:file`: The schema file
			- must have header
	@output:
		`dsn`: The dsn
	@args:
		`delimit`:The delimit of input file. Default: `\\t`
	"""
	pImportData                  = Proc(desc = 'Create tables and import the data')
	pImportData.input            = "dsn, datafile:file"
	pImportData.output           = "dsn:var:{{i.dsn}}"
	pImportData.args.delimit     = "\t"
	pImportData.args.drop        = False
	pImportData.envs.dsnparse    = sql.dsnparse
	pImportData.envs.schemaparse = sql.schemaparse
	pImportData.envs.tablename   = lambda fn: __import__('os').path.basename(fn).split('.')[0]
	pImportData.lang             = params.python.value
	pImportData.script           = 'file:scripts/sql/pImportData.py'
	return pImportData

@procfactory
def _pUpdateTable():
	"""
	@name:
		pUpdateTable
	@description:
		Update table using sql.
	@input:
		`dsn`: The dsn to connect to the database
			- currently support `sqlite:file=...`
	@output:
		`dsn`: The dsn
	@args:
		`sql`: The sql to update the table (list)
	"""
	pUpdateTable               = Proc(desc = 'Update table using sql.')
	pUpdateTable.input         = 'dsn'
	pUpdateTable.output        = 'dsn:var:{{i.dsn}}'
	pUpdateTable.args.sql      = []
	pUpdateTable.envs.dsnparse = sql.dsnparse
	pUpdateTable.lang          = params.python.value
	pUpdateTable.script        = 'file:scripts/sql/pUpdateTable.py'
	return pUpdateTable

@procfactory
def _pSelectTable():
	"""
	@name:
		pSelectTable
	@description:
		Select data from table and dump it.
	@input:
		`dsn`: The dsn to connect to the database
			- currently support `sqlite:file=...`
	@output:
		`outfile:file`: The dumped file
	@args:
		`sql`: The sql to select data from the table (list)
	"""
	pSelectTable               = Proc(desc = 'Select data from table and dump it.')
	pSelectTable.input         = 'dsn'
	pSelectTable.output        = 'outfile:file:sqlSelectTable-{{i.dsn, args.sql | lambda x,y: __import__("hashlib").md5(str(x) + str(y)).hexdigest()[:16]}}.dumped.txt'
	pSelectTable.args.sql      = []
	pSelectTable.envs.dsnparse = sql.dsnparse
	pSelectTable.lang          = params.python.value
	pSelectTable.script        = 'file:scripts/sql/pSelectTable.py'
	return pSelectTable





from bioprocs.utils import shell2 as shell

infile   = {{i.infile | quote}}
config   = {{i.config | quote}}
outdir   = {{o.outdir | quote}}
Rscript  = {{args.Rscript | quote}}
formula  = {{args.formula | repr}}
title    = {{args.title | repr}}
figtype  = {{args.figtype | repr}}
passed   = {{args.passed | repr}}
region   = {{args.region | repr}}
regfile  = {{args.regfile | repr}}
macro    = {{args.macro | repr}}
ggs      = {{args.ggs | repr}}
devpars  = {{args.devpars | repr}}
vcfstats = {{args.vcfstats | repr}}

shell.load_config(vcfstats = vcfstats)

shell.fg.vcfstats(
	vcf      = infile,
	outdir   = outdir,
	formula  = formula,
	title    = title,
	loglevel = 'DEBUG',
	figtype  = figtype or False,
	region   = region or False,
	Region   = regfile or False,
	passed   = passed,
	macro    = macro or False,
	ggs      = ggs or False,
	devpars  = devpars or False,
	config   = config or False
)



"""Commands of bcftools"""

from modkit import Modkit
from pyppl import Box, Proc
from . import params, delefactory, procfactory
Modkit().delegate(delefactory())

@procfactory
def _pQuery():
	"""
	@input
		infile: The input VCF file. Currently only single VCF file is supported.
	@output
		outfile: The output file
	@args:
		bcftools (str): Path to bcftools.
		params (Box): Other parameters for `bcftools query`.
			- See: https://samtools.github.io/bcftools/bcftools.html#query
	"""
	return Box(
		desc   = "Extracts fields from VCF or BCF files and outputs them in user-defined format",
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | stem | stem}}.query.txt',
		lang   = params.python.value,
		args   = Box(
			bcftools = params.bcftools.value,
			params = Box()
		)
	)

@procfactory
def _pView():
	"""
	@input
		infile: The input VCF file. Currently only single VCF file is supported.
		samfile: The sample name file to extract samples from infile
			- It also can be samples directly, separated by comma
			- If this is provided, `args.params.s` and `args.params.S` will be ignored
			- See https://samtools.github.io/bcftools/bcftools.html#view
	@output
		outfile: The output file
	@args:
		tabix    (str) : Path to tabix.
		bcftools (str) : Path to bcftools.
		gz       (bool): Whether output gzipped vcf file or not.
		nthread  (int) : Number of threads to use.
		params   (Box) : Other parameters for `bcftools view`.
			- See: https://samtools.github.io/bcftools/bcftools.html#view
	"""
	return Box(
		desc   = "View, subset and filter VCF or BCF files by position and filtering expression.",
		input  = 'infile:file, samfile:var',
		output = 'outfile:file:{{i.infile | stem | stem}}.vcf{{".gz" if args.gz else ""}}',
		lang   = params.python.value,
		args   = Box(
			gz       = False,
			nthread  = 1,
			tabix    = params.tabix.value,
			bcftools = params.bcftools.value,
			params   = Box()
		)
	)

@procfactory
def _pReheader():
	"""
	@input:
		infile: The input VCF file
		hfile: The new header file
		samfile: The new sample name file
			- It also can be samples directly, separated by comma
			- If this is provided, `args.params.samples` will be ignored
			- See https://samtools.github.io/bcftools/bcftools.html#reheader
	@output:
		outfile: The output VCF file with new header.
	@args:
		bcftools (str): Path to bcftools.
		nthread  (int): Number of threads to use.
		params   (Box): Other parameters for `bcftools view`.
			- See https://samtools.github.io/bcftools/bcftools.html#reheader
	"""
	return Box(
		desc   = "Modify header of VCF/BCF files, change sample names",
		input  = 'infile:file, hfile:file, samfile:var',
		output = 'outfile:file:{{i.infile | bn}}',
		lang   = params.python.value,
		args   = Box(
			bcftools = params.bcftools.value,
			nthread  = 1,
			params   = Box()
		)
	)

@procfactory
def _pFilter():
	"""
	@input:
		infile: The input VCF file
	@output:
		outfile: The output filtered VCF file
		statfile: Statistics of variants for each FILTER.
	@args:
		bcftools (str) : Path to bcftools
		nthread  (int) : Number of threads to use.
		gz       (bool): Whether output gzipped vcf file or not.
			- Overwrite `args.params.O`
		keep     (bool): Whether should we keep the excluded variants or not.
			- If not, args about filter names will not work.
		include  (str|list|dict): include only sites for which EXPRESSION is true.
			- See: https://samtools.github.io/bcftools/bcftools.html#expressions
			- Use `args.params.include` or `args.params.exclude` only if you just have one filter.
			- If provided, `args.params.include/exclude` will be ignored.
			- If `str`/`list` used, The filter names will be `Filter%d`
			- A dict is used when keys are filter names and values are expressions
		exclude  (str|list|dict): exclude sites for which EXPRESSION is true.
			- See also `args.include`
		params   (Box)          : Other parameters for `bcftools filter`
			- See: https://samtools.github.io/bcftools/bcftools.html#filter
	"""
	return Box(
		desc   = "Apply fixed-threshold filters to VCF files",
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | bn}}, statfile:file:{{i.infile | stem | stem}}.filterstats.txt',
		lang   = params.python.value,
		args   = Box(
			bcftools = params.bcftools.value,
			nthread  = 1,
			gz       = False,
			stat     = False,
			keep     = True,
			include  = None,
			exclude  = None,
			params   = Box(
				mode = '+' # accumulating Filter names instead of repolacing
			)
		)
	)

@procfactory
def _pAnnotate():
	"""
	@input:
		infile: The input VCF file
	@output:
		outfile: The annotated output VCF file
	@args:
		tabix   (str)  : Path to tabix, used to index `annfile`
		bcftools (str) : Path to bcftools
		annfile  (file): The annotation file.
			- See: https://samtools.github.io/bcftools/bcftools.html#annotate
		nthread (int)     : Number of threads to use.
		cols    (str|list): Overwrite `-c/--columns`.
		header  (str|list): headers to be added.
		params  (Box)     : Other parameters for `bcftools annotate`
	"""
	return Box(
		desc   = 'Add or remove annotations from VCF files',
		lang   = params.python.value,
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | bn}}',
		args   = Box(
			tabix    = params.tabix.value,
			bcftools = params.bcftools.value,
			nthread  = 1,
			annfile  = '',
			cols     = [],
			header   = [],
			params   = Box()
		)
	)




from os import path
from glob import glob
from bioprocs.utils import runcmd, cmdargs
from bioprocs.utils.tsvio import TsvReader, TsvWriter

# plink -bfile x --recode A-transpose --out x.txt 
# x.txt.traw

indir   = {{i.indir | quote}}
outfile = {{o.outfile | quote}}
plink   = {{args.plink | quote}}
samid   = {{args.samid | quote}}
snpid   = {{args.snpid | quote}}
addchr  = {{args.addchr | repr}}
nors    = {{args.nors | quote}}
chroms  = {{args.chroms | repr}}

bedfile = glob(path.join(indir, '*.bed'))[0]
input   = path.splitext(bedfile)[0]
output  = path.splitext(outfile)[0]

params = {
	'bfile' : input,
	'recode': 'A-transpose',
	'out'   : output
}

cmd = '%s %s 1>&2' % (plink, cmdargs(params, equal = ' '))
runcmd(cmd)

fams = TsvReader(input + '.fam', ftype = 'nometa', delimit = ' ', head = False)
if samid == 'fid':
	header = "\t" + "\t".join(fams.dump(0)) + "\n"
elif samid == 'iid':
	header = "\t" + "\t".join(fams.dump(1)) + "\n"
else:
	header = "\t" + "\t".join(r[0] + '_' + r[1] for r in fams) + "\n"
fams.close()

gts = TsvReader(output + '.traw', ftype = 'nometa', skip = 1, head = False)

with open(outfile, 'w') as fout:
	fout.write(header)
	for gtline in gts:
		if snpid == 'raw':
			gtline[0] = gtline[1]
		else:
			if gtline[0] in chroms:
				gtline[0] = chroms[gtline[0]]
			elif 'chr' + gtline[0] in chroms:
				gtline[0] = chroms['chr' + gtline[0]]
			if addchr and not gtline[0].startswith('chr'):
				gtline[0] = 'chr' + gtline[0]
			gtline[0] = snpid.format(
				chr = gtline[0],
				pos = gtline[3],
				rs  = gtline[1] if 'rs' in gtline[1] else nors,
				ref = gtline[4],
				alt = gtline[5]
			)
		del gtline[1:6]
		fout.write('\t'.join(gtline) + '\n')





import os #, shutil
from subprocess import check_output

{{downloadCurl}}
{{runcmd}}
{{txtFilter}}
{{txtTransform}}

url = {{args.urls | json}}["{{i.in}}"]
tmpdir = "{{job.outdir}}/tmp"
if not os.path.exists(tmpdir):
	os.makedirs(tmpdir)
	
downfile = os.path.join(tmpdir, 'downloaded')
downloadCurl(url, downfile, {{args.username | quote}}, {{args.password | quote}}, {{args.curl | quote}})

# determin file types
output = check_output(['file', downfile])
if 'gzip' in output:
	ugfile = downfile + '.ungz'
	with open(ugfile, 'w') as f:
		f.write(check_output(['gunzip', downfile, '-c']))
	downfile = ugfile
elif 'Zip' in output:
	zipdir = os.path.join(tmpdir, '_unzipped')
	import zipfile, glob
	zipref = zipfile.ZipFile(downfile, 'r')
	zipref.extractall(zipdir)
	zipref.close()
	downfile = glob.glob(os.path.join(zipdir, '*'))[0]

outfile = "{{o.outfile}}"[:-3] if {{args.gz | lambda x: bool(x)}} else "{{o.outfile}}"
txtFilter(downfile, outfile + '.filtered', {{args.cols | lambda x: [] if not x else __import__('json').dumps(x) if isinstance(x, list) else '"'+ x +'"'}}, {{args.rowfilter | lambda x: 'False' if not x else x}}, {{args.header}}, {{args.skip}}, {{args.delimit | quote}})

{% if i['in'] == 'KEGG_2016_gmt' %}
def transformer(parts):
	return [ p.replace('/', '-') if i == 0 else p[:-4] if p.endswith(',1.0') else p for i,p in enumerate(parts) ]

tmpfile = outfile + '.kegg'
txtTransform(outfile + '.filtered', tmpfile, transform = transformer, header = {{args.header}}, delimit = {{args.delimit | quote}})
os.rename(tmpfile, outfile + '.filtered')
{% endif %}

{% if args.transform %}
txtTransform(outfile + '.filtered', outfile, transform = {{args.transform}}, header = {{args.header}}, delimit = {{args.delimit | quote}})
{% else %}
os.rename(outfile + '.filtered', outfile)
{% endif %}

if {{args.gz | bool}}:
	runcmd ('gz "%s"' % outfile)

#shutil.rmtree(tmpdir)


# sqlite:file=/path/to/sqlite.db;...
{{dsnparse}}
{{schemaparse}}

dsn = dsnparse({{i.dsn | quote}})
if not hasattr(dsn, 'scheme'):
	raise Exception('Cannot determine the scheme from DSN string.')

name, fields = schemaparse({{i.schema | quote}}, type = {{args.intype | quote}}, scheme = dsn.scheme, delimit = {{args.delimit | quote}})

if dsn.scheme in ['sqlite', 'sqlite3']:
	assert hasattr(dsn, 'file') and dsn.file != ':memory:'
	import sqlite3
	conn = sqlite3.connect(dsn.file)
	{# Drop the existing table first #}
	{% if args.drop %}
	dropsql = 'DROP TABLE IF EXISTS "%s"' % name
	conn.execute(dropsql)
	{% endif %}

	createsql  = 'CREATE TABLE IF NOT EXISTS "%s" (\n' % name
	for i, field in enumerate(fields):
		fname, ftype, fstat = field
		if i == len(fields) - 1:
			createsql += '  "%s" %s %s\n' % (fname, ftype, fstat)
		else:
			createsql += '  "%s" %s %s,\n' % (fname, ftype, fstat)
	createsql += ') %s' % dsn.after

	stderr.write('Executing SQL:\n')
	stderr.write(createsql + '\n')
	conn.execute(createsql)
	conn.commit()
	conn.close()
else:
	raise NotImplementedError('Only support sqlite3 currently.')





from bioprocs.utils import logger, runcmd, cmdargs
from pyppl import Box

params = {{args.params}}

qcdir = {{i.qcdir | quote}}
try:
	{% if args.tool == 'multiqc' %}
	params['o'] = {{o.outdir | quote}}
	params['p'] = True
	cmd = '{{args.multiqc}} "{{i.qcdir}}" %s' % cmdargs(params)
	runcmd(cmd)

	{% else %}
	raise Exception('Tool {{args.tool}} not supported.')

	{% endif %}
except Exception as ex:
	logger.error("Job failed: %s" % str(ex))
	raise



from os import path
from bioprocs.utils import logger

infiles = {{i.infiles | repr}}
outfile = {{o.outfile | quote}}
rsmap   = {{args.rsmap | : None if not a else a | repr}}
fn2sam  = {{args.fn2sample}}
confcut = {{args.confcut | repr}}

def readone(infile, getsnps = True, progress = ''):
	logger.info('Reading ' + progress + ': ' + path.basename(infile) + ' ...')
	gts        = []
	snps       = []
	gts_append = gts.append
	if getsnps:
		snps_append = snps.append
	with open(infile) as fin:
		fin.readline()
		fin.readline()
		for line in fin:
			line = line.strip()
			if not line: continue
			snp, gt, conf = line.split()[:3]
			if confcut and float(conf) >= confcut:
				gt = 'NA'
			if getsnps:
				snps_append(snp)
			gts_append(gt)
	return snps, gts

ret     = None
samples = ['Variant']
inlen   = len(infiles)
for i, infile in enumerate(infiles):
	sam = path.splitext(path.basename(infile))[0]
	if fn2sam:
		samples.append(fn2sam(sam))
	else:
		samples.append(sam)
	snps, gts = readone(infile, i == 0, progress = '{}/{}'.format(i+1, inlen))
	if snps: 
		ret = [(snps[i], [gt]) for i, gt in enumerate(gts)]
	else:
		[ret[i][1].append(gt) for i, gt in enumerate(gts)]
		
if rsmap:
	logger.info('Reading rs mapping file ...')
	with open(rsmap) as fin:
		snp2rs = {snp:rs for line in fin if line.strip() for rs, snp in [line.split()[:2]]}
	ret = [(snp2rs[r[0]], r[1]) for r in ret if snp2rs[r[0]].startswith('rs')]

logger.info('Saving results ...')
with open(outfile, 'w') as fout:
	fout.write('\t'.join(samples) + '\n')
	for rs, gts in ret:
		fout.write(rs + '\t' + '\t'.join(gts) + '\n')



"""Processes related to DNA/protein sequences"""
from pyppl import Proc, Box
#from .utils import runcmd, helpers, genenorm, write
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pConsvPerm():
	"""
	@name:
		pConsvPerm
	@description:
		Generate a null distribution of conservation scores.
	@input:
		`seed`: The seed to generate the random regions. Default: None
	@output:
		`outfile:file`: A file with mean conservation scores sorted descendingly.
	@args:
		`len`: The length of a random region. Default: 50
		`nperm`: Number of permutations. Default: 1000
		`gsize`: The chrom size file.
		`bedtools`: The path of bedtools.
		`bwtool`: The path of bwtool.
		`consvdir`: The directory containing bigwig files of conservation scores
			- The bigwig file should start with chr name: chrN.*
	@requires:
		[bwtool](https://github.com/CRG-Barcelona/bwtool)
		[bedtools](http://bedtools.readthedocs.io/en/latest/content/bedtools-suite.html)
	"""
	pConsvPerm               = Proc(desc = 'Generate a null distribution of conservation scores.')
	pConsvPerm.input         = 'seed'
	pConsvPerm.output        = 'outfile:file:consv-len{{args.len}}-nperm{{args.nperm}}-{{i.seed}}.txt'
	pConsvPerm.args.len      = 50
	pConsvPerm.args.nperm    = 1000
	pConsvPerm.args.consvdir = params.consvdir.value
	pConsvPerm.args.gsize    = params.gsize.value
	pConsvPerm.args.bedtools = params.bedtools.value
	pConsvPerm.args.bwtool   = params.bwtool.value
	pConsvPerm.args.seed     = None
	pConsvPerm.lang          = params.python.value
	pConsvPerm.script        = "file:scripts/seq/pConsvPerm.py"
	return pConsvPerm


@procfactory
def _pConsv():
	"""
	@name:
		pConsv
	@description:
		Get the conservation scores of regions.
		It uses wigFix to find the conservation scores.
		But first you have to convert those wigFix.gz files to bigWig files using ucsc-wigToBigWig
	@input:
		`bedfile:file`: The bedfile with regions in the same chromosome
		`permfile:file`:The permutaiton file generated by `pConsvPerm`, used to calculate p-values
	@output:
		`outfile:file`: The output file
	@args:
		`consvdir`:   The bigwig directory, the bigwig files must be named as "chrN.*.bw"
			- For example: `chr1.phyloP30way.bw`
		`bwtool`:   The path of bwtool executable. Default: `bwtool`
		`bedtools`: The path of bedtools executable. Default: `bedtools`
		`pval`:     Whether calculate pvalue of each region. Default: False
			- In this case, the `i.permfile` can be ignored.
	@requires:
		[bwtool](https://github.com/CRG-Barcelona/bwtool)
		[bedtools](http://bedtools.readthedocs.io/en/latest/content/bedtools-suite.html)
	"""
	pConsv               = Proc(desc = 'Get the conservation scores of regions.')
	pConsv.input         = "bedfile:file, permfile:file"
	pConsv.output        = "outfile:file:{{i.bedfile | fn}}-consv.bed"
	pConsv.args.bwtool   = params.bwtool.value
	pConsv.args.consvdir = params.consvdir.value
	pConsv.args.pval     = False
	pConsv.lang          = params.python.value
	pConsv.script        = "file:scripts/seq/pConsv.py"
	return pConsv


@procfactory
def _pPromoters():
	"""
	@name:
		pPromoters
	@description:
		Get the promoter regions in bed format of a gene list give in infile.
		Gene names are supposed to be normalized by `gene.pGeneNameNorm`
	@input:
		`infile:file`: the gene list file
	@output:
		`outfile:file`: the bed file containing the promoter region
	@args:
		`region`: The region to output. Default: `Box(up = 2000, down = None, withbody = False)`
			- `up`: The upstream distance to TSS.
			- `down`: The downstream distance to TSS. Defaults to `args.region.up` if `None`
			- `withbody`: Include gene body in the output region? Default: `False`
		`notfound`: How to deal with records can't be found. Default: `skip`
			- `skip` : Skip the record
			- `error`: Report error and exit
		`refgene`: The ref gene file. Default: `@params.refgene`
		`inopts` : The options for input file. Default: `Box(cnames = False, genecol = 0, delimit = "\t")`
			- `cnames`: Whether the input file has header
			- `genecol`: The 0-based index or the colname of gene column.
			- `delimit`: The delimit of the input file.
		`outopts`:  The options for output file. Default: `Box(cnames = False)`
	"""
	pPromoters               = Proc(desc = 'Get the promoter regions in bed format of a gene list give in infile.')
	pPromoters.input         = "infile:file"
	pPromoters.output        = "outfile:file:{{i.infile | fn}}-promoters.bed"
	pPromoters.args.region   = Box(up = 2000, down = None, withbody = False)
	pPromoters.args.notfound = 'skip' # error
	pPromoters.args.inopts   = Box(cnames = False, delimit = "\t")
	pPromoters.args.genecol  = 0
	pPromoters.args.refgene  = params.refgene.value
	pPromoters.lang          = params.python.value
	pPromoters.script        = "file:scripts/seq/pPromoters.py"
	return pPromoters




from pyppl import Box
from os import path
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter
from bioprocs.utils import shell2 as shell, logger

infile   = {{i.infile | quote}}
outdir   = {{o.outdir | quote}}
dot      = {{args.dot | quote}}
lichee   = {{args.lichee |quote}}
params   = {{args.params | repr}}
fishplot = {{args.fishplot |repr}}
Rscript  = {{args.Rscript |quote}}
devpars  = {{args.devpars |repr}}

shell.load_config(lichee = {'_exe': lichee, '_prefix': '-'}, Rscript = Rscript, dot = dot)

# expected input:
# #chr    position    description    Normal    S1    S2    S3    S4
# 17      123456      A/T DUSP19     0.0       0.1   0.2   0.25  0.15
# 11      341567      C/G MUC16      0.0       0.4   0.09  0.38  0.24
# 9       787834      A/C OR2A14     0.0       0.35  0.14  0.17  0.48

# let's see if the input file is pyclone directory
if path.isdir(infile) and infile.endswith('.pyclone'):
	# tables/loci.tsv
	#mutation_id	sample_id	cluster_id	cellular_prevalence	cellular_prevalence_std	variant_allele_frequency	gene
	#chr10:17875816	NA12878	2	0.828865827591	0.0355432431918	0.904883227176	Gene1
	#chr10:17875816	NA12156	2	0.741264908504	0.11050033704	0.560311284047	Gene2

	indata = {} # {mutation: {gene: gene, sample1: {cluster_id: cluster_id, cp = cp}, sample2: ...}}
	reader = TsvReader(path.join(infile, 'tables/loci.tsv'))
	samples = set()
	for i, r in enumerate(reader):
		indata.setdefault(r.mutation_id, Box())
		if not indata[r.mutation_id].get('_gene'):
			indata[r.mutation_id]._gene = r.gene
		indata[r.mutation_id]._cluster     = r.cluster_id
		indata[r.mutation_id][r.sample_id] = r.cellular_prevalence
		samples.add(r.sample_id)

	profile = lambda vafs: ''.join(str(int(bool(vaf))) for vaf in vafs)

	samples = list(samples)
	writer = TsvWriter(path.join(outdir, 'input.txt'))
	writer.cnames = ['chr', 'position', 'description', 'Normal'] + samples
	writer.writeHead(lambda cnames: '#' + '\t'.join(cnames))
	clusters = {}
	i = 1
	for mutation, info in indata.items():
		vafs = [0.0] + [info[sample] for sample in samples]
		prof = profile(vafs)
		clusters.setdefault(info.cluster, [])
		clusters[info.cluster].append(Box(
			profile = prof,
			vafs    = vafs,
			index   = str(i)
		))
		i += 1
		writer.write(mutation.split(':') + [info._gene] + vafs)
	writer.close()

	# 01111     0.0   0.1   0.23  0.23  0.13    1,2
	# 01111     0.0   0.4   0.4   0.45  0.45    3
	# 01001     0.0   0.4   0.0   0.0   0.24    4
	writer = TsvWriter(path.join(outdir, 'clusters.txt'))
	for cluster_id, info in clusters.items():
		writer.write([info[0].profile] + [
			sum(float(inf.vafs[i]) for inf in info)/float(len(info)) for i in range(len(info[0].vafs))
		] + [','.join(inf.index for inf in info)])
	writer.close()
	params.cp = True
	infile = path.join(outdir, 'input.txt')
	params.clustersFile = path.join(outdir, 'clusters.txt')
elif infile.endswith('.maf') or infile.endswith('.maf.gz'):
	reader = TsvReader(infile)
	if 't_alt_count' not in reader.cnames:
		raise ValueError('t_alt_count not found in MAF file.')
	if 't_ref_count' not in reader.cnames:
		raise ValueError('t_ref_count not found in MAF file.')

	# get all mutations
	mutations = {}
	samples = set()
	for r in reader:
		mut = r.Chromosome + ':' + r.Start_Position
		mutations.setdefault(mut, Box())
		try:
			mutations[mut][r.Tumor_Sample_Barcode] = float(r.t_alt_count) / (float(r.t_alt_count) + float(r.t_ref_count))
		except ZeroDivisionError:
			logger.warning(
				f'Failed to get allele frequency for variant: {mut} with alt/ref counts: {r.t_alt_count}, {r.t_ref_count}, setting VAF to 0')
			mutations[mut][r.Tumor_Sample_Barcode] = 0.0
		mutations[mut]._gene = r.Hugo_Symbol

		samples.add(r.Tumor_Sample_Barcode)
	samples = list(samples)

	writer = TsvWriter(path.join(outdir, 'input.txt'))
	writer.cnames = ['chr', 'position', 'description', 'Normal'] + samples
	writer.writeHead(lambda cnames: '#' + '\t'.join(cnames))
	for mut, info in mutations.items():
		writer.write(mut.split(':') + [info._gene, 0.0] + [
			info.get(sample, 0.0) for sample in samples
		])
	writer.close()
	infile = path.join(outdir, 'input.txt')
else:
	shell.ln_s(infile, path.join(outdir, path.basename(infile)))

prefix = path.join(outdir, path.basename(infile).split('.')[0])
outfile = prefix + '.trees.txt'

params.i = infile
params.n = 0
params.o = outfile
params.color = True
params.v = True
params.dot = True
params.dotFile = prefix + '.dot'

shell.fg.lichee('-build', **params)

shell.fg.dot(params.dotFile, f'-Gdpi={devpars.res}', T = 'png', o = prefix + '.png')

# save clusters and snv infos in different file for reporting
snvfile = prefix + '.snvinfo.txt'
snvinfos = {}
nodes_started = False
info_started = False
with open(outfile, 'r') as fout:
	for line in fout:
		line = line.strip()
		if line == 'Nodes:':
			nodes_started = True
			continue
		if line == 'SNV info:':
			info_started = True
			continue
		if not line:
			nodes_started = False
			info_started = False
			continue
		if nodes_started:
			parts = line.split('\t')
			cluster = int(parts[0]) - 1
			snvs = parts[3:]
			for snv in snvs:
				snvinfos.setdefault(snv, {})
				snvinfos[snv]['cluster'] = cluster
			continue
		if info_started:
			snv, info = line.split(': ', 1)
			snvinfos[snv]['info'] = info
with open(snvfile, 'w') as fsnv:
	fsnv.write('SNV\tClusterID\tDescription\n')
	for snv, info in snvinfos.items():
		fsnv.write(f'{snv}\t{info["cluster"]}\t{info["info"]}\n')




from pyppl import Box
from bioprocs.utils import ensureBox, shell2 as shell

args = {{i.args if isinstance(i.args, str) else repr(i.args)}}
args = ensureBox(args)

cmd = {{args.cmd | repr}}

if not cmd:
	raise ValueError('No cmd (args.cmd) specified')

if isinstance(cmd, dict):
	cmd, exe = cmd.items()[0]
elif isinstance(cmd, (tuple, list)):
	cmd, exe = cmd
else:
	cmd, exe = cmd, cmd

cmds    = cmd.split('.')
command = shell.fg
for cmd in cmds:
	command = getattr(command, cmd)
command(_exe = exe, **args)



from pyppl import Box
from bioprocs.utils.tsvio2 import TsvWriter, TsvReader, TsvRecord

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
inopts  = {{args.inopts | repr}}
outopts = {{args.outopts | repr}}
helper  = {{args.helper | repr}}
if not isinstance(helper, list):
	helper = [helper]

helper = [line for line in helper if line]
exec('\n'.join(helper), globals())

row = {{args.row}}
inopts['row'] = {{args.inopts.get('row', None)}}

reader       = TsvReader(infile, **inopts)
writer       = TsvWriter(outfile, delimit = outopts.get('delimit', "\t"))
outcnames    = outopts.get('cnames', True)
headCallback = {{args.outopts.get('headCallback', True)}}
if outcnames is True:
	writer.cnames = reader.cnames
	writer.writeHead(headCallback)
elif isinstance(outcnames, (list, tuple)):
	writer.cnames = list(outcnames)
	writer.writeHead(headCallback)

for r in reader:
	rec = row(r)
	if rec is False:
		continue
	if rec is None or rec is True:
		rec = r
	writer.write(rec)
writer.close()




colnames = []
rownames = set()
data     = {}
with open('{{i.infile}}') as f:
	for line in f:
		line  = line.strip()
		if not line: continue
		parts = line.split('\t')
		rname = parts.pop(0)
		colnames.append(rname)
		parts.pop(0) # annotation
		data[rname] = parts
		rownames = rownames | set(parts)

with open('{{o.outfile}}', 'w') as f:
	f.write("\t" + "\t".join(colnames) + "\n")
	for rname in rownames:
		tmp = ['1' if rname in data[cname] else '0' for cname in colnames]
		tmp.insert(0, rname)
		f.write("\t".join(tmp) + "\n")



"""Snp-array utilities"""
from pyppl import Proc, Box
from . import params
#from .utils import runcmd, helpers
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pGistic():
	"""
	@name:
		pGistic
	@description:
		Runing GISTIC to get CNV results.
		see: ftp://ftp.broadinstitute.org/pub/GISTIC2.0/GISTICDocumentation_standalone.htm
	@input:
		`segfile:file`: Segmentation File
		`mkfile:file` : Markers File
		`alfile:file` : Array List File
		`cnvfile:file`: CNV File
	@output:
		`outdir:dir`: The output directory
			- All Lesions File (all_lesions.conf_XX.txt, where XX is the confidence level)
			- Amplification Genes File (amp_genes.conf_XX.txt, where XX is the confidence level)
			- Deletion Genes File (del_genes.conf_XX.txt, where XX is the confidence level)
			- Gistic Scores File (scores.gistic)
			- Segmented Copy Number (raw_copy_number.pdf)
	@args:
		`gistic`: The path to gistic.
		`genome`: The genome used to select refgene file from refgenefiles.
		`mcr`:    The mcr path
		`params`: Other params for gistic
	"""
	pGistic             = Proc(desc = 'Runing GISTIC to get CNV results.')
	pGistic.input       = 'segfile:file, mkfile:file, alfile:file, cnvfile:file'
	pGistic.output      = 'outdir:dir:{{i.segfile | fn}}.gistic'
	pGistic.args.gistic = params.gistic.value
	pGistic.args.genome = params.genome.value
	pGistic.args.mcr    = params.mcr # 2.0 requires r2014a
	pGistic.args.params = Box()
	pGistic.lang        = params.python.value
	pGistic.script      = "file:scripts/snparray/pGistic.py"
	return pGistic

@procfactory
def _pSNP6Genotype():
	"""
	@name:
		pSNP6Genotype
	@description:
		Call genotypes from GenomeWideSNP_6 CEL file
	@input:
		`celfile:file`: the CEL file
	@output:
		`outfile:file`: the outfile containing probe name and genotypes
		- format: `<Probe name>\t<genotype>`
		- `<genotype>` = 0: AA, 1: AB, 2: BB
	@requires:
		[bioconductor-crlmm](http://bioconductor.org/packages/release/bioc/html/crlmm.html)
	"""
	pSNP6Genotype        = Proc()
	pSNP6Genotype.input  = "celfile:file"
	pSNP6Genotype.output = "outfile:file:{{celfile | fn}}.geno.txt"
	pSNP6Genotype.lang   = "Rscript"
	pSNP6Genotype.script = """
	require(oligoClasses)
	library(crlmm)
	crlmmResult <- crlmm("{{celfile}}", SNRMin=0)
	gts = calls(crlmmResult) - 1
	#outfile = paste (sep="/", "/data2/junwenwang/panwen/output/TCGA-genotypes/LUAD/gts", paste(basename(args[1]), "gts", sep="."))
	write.table(gts, file="{{outfile}}", sep="\\t", row.names=TRUE, quote=FALSE, col.names=FALSE)
	"""
	return pSNP6Genotype

@procfactory
def _pGenoToAvInput():
	"""
	@name:
		pGenoToAvInput
	@description:
		Convert the genotype called by pSNP6Genotype to [ANNOVAR input file](http://annovar.openbioinformatics.org/en/latest/user-guide/input/#annovar-input-file) using dbSNP identifiers.
	@input:
		`genofile:file`: the genofile generated by pSNP6Genotype, must be sorted by probe names
		`annofile:flie`: the annotation file downloaded from http://www.affymetrix.com/support/technical/annotationfilesmai.affx
			- Could be in .gz format
	@output:
		`outfile:file`: the avinput file
	@requires:
		[python-read2](https://github.com/pwwang/read2)
	"""
	pGenoToAvInput = Proc()
	pGenoToAvInput.input  = "genofile:file, annofile:file"
	pGenoToAvInput.output = "outfile:file:{{genofile | fn}}.avinput"
	pGenoToAvInput.script = """
	#!/usr/bin/env python
	from read2 import read2
	fout = open ("{{outfile}}", "w")
	def rmatch (line1, line2):
		line2 = [item[1:-1] for item in line2]
		if line1[0].startswith ("#"): return -1
		if line2[0].startswith ("#") or line2[0].startswith("Probe Set ID"): return 1
		tomatch = line2[0]
		if line1[0] < tomatch: return -1
		if line1[0] > tomatch: return 1
		return 0

	def ract (line1, line2):
		line2 = [item[1:-1] for item in line2]
		snp   = line2[1]
		chr   = line2[2]
		pos   = line2[3]
		allA  = line2[8]
		allB  = line2[9]
		comm  = line1
		comm.pop(0)
		fout.write (" ".join([chr,pos,pos,allA, allB, snp, "|".join(comm)]) + "\\n")
	r = read2 ("{{genofile}}", "{{annofile}}")
	r.delimit ("\\t", ",")
	r.match (rmatch)
	r.act (ract)
	r.run()
	fout.close()
	"""
	return pGenoToAvInput




from collections import defaultdict
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
tmbtype = {{args.type | quote}}

"""
Red --> Nonsense_Mutation, frameshift_variant, stop_gained, splice_acceptor_variant, splice_acceptor_variant&intron_variant, splice_donor_variant, splice_donor_variant&intron_variant, Splice_Site, Frame_Shift_Del, Frame_Shift_Ins

Blue --> splice_region_variant, splice_region_variant&intron_variant, missense, non_coding_exon_variant, missense_variant, Missense_Mutation, exon_variant, RNA, Indel, start_lost, start_gained, De_novo_Start_OutOfFrame, Translation_Start_Site, De_novo_Start_InFrame, stop_lost, Nonstop_Mutation, initiator_codon_variant, 5_prime_UTR_premature_start_codon_gain_variant, disruptive_inframe_deletion, inframe_deletion, inframe_insertion, In_Frame_Del, In_Frame_Ins

Green --> synonymous_variant, 5_prime_UTR_variant, 3_prime_UTR_variant, 5'Flank, 3'Flank, 3'UTR, 5'UTR, Silent, stop_retained_variant

Orange --> others, SV, upstreamgenevariant, downstream_gene_variant, intron_variant, intergenic_region
"""
# oncotator MAFs
INTRON = "Intron"
FIVE_PRIME_UTR = "5'UTR"
THREE_PRIME_UTR = "3'UTR"
IGR = "IGR"
FIVE_PRIME_PRIME_FLANK = "5'Flank"
THREE_PRIME_PRIME_FLANK = "3'Flank"
MISSENSE = "Missense_Mutation"
NONSENSE = "Nonsense_Mutation"
NONSTOP = "Nonstop_Mutation"
SILENT = "Silent"
SPLICE_SITE = "Splice_Site"
IN_FRAME_DEL = "In_Frame_Del"
IN_FRAME_INS = "In_Frame_Ins"
FRAME_SHIFT_INS = "Frame_Shift_Ins"
FRAME_SHIFT_DEL = "Frame_Shift_Del"
START_CODON_SNP = "Start_Codon_SNP"
START_CODON_INS = "Start_Codon_Ins"
START_CODON_DEL = "Start_Codon_Del"
STOP_CODON_INS = "Stop_Codon_Ins"
STOP_CODON_DEL = "Stop_Codon_Del"
# Note: A STOP_CODON_SNP is a nonstop mutation (or Silent)
DE_NOVO_START_IN_FRAME = "De_novo_Start_InFrame"
DE_NOVO_START_OUT_FRAME = "De_novo_Start_OutOfFrame"
RNA = "RNA"
LINCRNA = "lincRNA"
# vcf2maf
TRANSLATION_START_SITE = 'Translation_Start_Site'
SPLICE_REGION = "Splice_Region"
TARGETED_REGION = "Targeted_Region"

def nonsyn(maf, outfile):
	# should only include Red and Blue
	excludes = (INTRON, FIVE_PRIME_UTR, THREE_PRIME_UTR, IGR, FIVE_PRIME_PRIME_FLANK, THREE_PRIME_PRIME_FLANK, SILENT, )
	reader = TsvReader(infile)
	# Hugo_Symbol     Entrez_Gene_Id  Center  NCBI_Build      Chromosome      Start_Position  End_Position    Strand  Variant_Classification  Variant_Type    Reference_Allele        Tumor_Seq_Allele1       Tumor_Seq_Allele2       dbSNP_RS      dbSNP_Val_Status        Tumor_Sample_Barcode ...
	tmb = defaultdict(lambda: 0)
	for r in reader:
		if r.Variant_Classification in excludes:
			continue
		tmb[r.Tumor_Sample_Barcode] += 1
	writer = TsvWriter(outfile)
	writer.cnames = ['Sample', 'TMB']
	writer.writeHead()
	for key, val in tmb.items():
		writer.write([key, val])

globals()[tmbtype](infile, outfile)


"""Script entry point for bioprocs"""
import sys
import copy
from pprint import pformat
from .. import params
from .utils import highlightMulti, Module, Pipeline, subtractDict
from .arguments import commands
from pyparam import HelpAssembler, Helps
from pyppl import Box
from pyppl.utils import config

HELP_ASSEMBLER = HelpAssembler()

def showParams(opts):
	"""Show and query a parameter"""
	# get is preserved for Box
	if opts['get']:
		for pname in opts._:
			if pname not in params:
				raise ValueError('Parameter %r is not defined.' % pname)
			print(params[pname].value)
	else:
		pminfo  = Helps()
		pminfo.add('params', sectype = 'option', prefix = '')
		queries = [query.lower() for query in opts._]
		for key in sorted(params._dict()):
			if key in params._hopts:
				continue
			for query in queries:
				param = params[key]
				if query not in param.str().lower() and query not in key.lower() \
					and not any(query in desc for desc in param.desc):
					continue
				pminfo['params'].addParam(param)

		print('\n'.join(highlightMulti(line, queries)
			for line in  HELP_ASSEMBLER.assemble(pminfo)), end = '')

def listProcs(opts):
	"""List processes"""
	if not opts._ and not opts.all: # list modules only
		pminfo  = Helps()
		pminfo.add('modules', sectype = 'option', prefix = '')
		for module in Module.modules():
			Module(module).toHelps(pminfo['modules'])
		print('\n'.join(HELP_ASSEMBLER.assemble(pminfo)), end = '')
	elif not opts._:  # list all modules and processes
		pminfo  = Helps()
		modules = Module.modules()
		nmods   = len(modules)
		for module in modules:
			Module(module).toHelpsWithProcs(pminfo, nmods)
		print('\n'.join(HELP_ASSEMBLER.assemble(pminfo)), end = '')
	else: # list specific modules and their processes
		pminfo  = Helps()
		modules = Module.modules()
		nmods   = len(opts._)
		for module in opts._:
			if module not in modules:
				raise ValueError('No such module: %s' % module)
			Module(module).toHelpsWithProcs(pminfo, nmods)
		print('\n'.join(HELP_ASSEMBLER.assemble(pminfo)), end ='')

def proc(opts):
	"""Query processes"""
	modules = Module.modules()
	nmods   = len(modules)
	helps   = Helps()
	for i, module in enumerate(modules):
		module = Module(module)
		if any(q.lower() in module.name.lower() for q in opts._):
			module.toHelpsWithProcs(helps, nmods)
		else:
			module.toHelpsAsSec(helps)
			with module.loadProcs("%s/%s" % (i+1, nmods)) as procs:
				for name, prc in procs.items():
					if any(q.lower() in name.lower() for q in opts._):
						prc.toHelps(helps.select(module.name + ':'))
	print('\n'.join(highlightMulti(line, opts._)
		for line in  HELP_ASSEMBLER.assemble(helps)), end = '')

def complete(opts):
	"""Shell completions"""
	from completions import Completions
	comp = Completions(desc = 'Bioprocs utilities.')
	# add builtin commands
	for cmd, pms in commands._cmds.items():
		if cmd == '_':
			continue
		comp.addCommand(cmd, pms._desc)
		pms._addToCompletions(comp.command(cmd), withtype = False, alias = True)

	# add pipelines
	for pipeline in Pipeline.pipelines():
		Pipeline(pipeline).addToCompletions(comp)

	# add modules/processes
	for module in Module.modules():
		Module(module).addToCompletions(comp)

	compcode = comp.generate(shell = opts.s, auto = opts.a)
	if not opts.a:
		print(compcode)

def profile(opts):
	"""List avaiable running profiles"""
	if opts._:
		config._load(opts._)

	base = { '_flowchart': {},
		'_log': {},
		'args': {},
		'sgeRunner': {},
		'sshRunner': {},
		'tplenvs': {}}

	helps = Helps()
	helps.add('Profile: "default"', sectype = 'plain')
	helps.select('Profile: "default"').add(
		pformat(subtractDict(config, base), indent = 2, width = 100))

	default = copy.deepcopy(config)
	for prof in config._profiles:
		if prof == 'default':
			continue
		helps.add('Profile: "%s"' % prof, sectype = 'plain')

		with config._with(prof, copy = True) as profconf:
			profconf = subtractDict(profconf, default)
			profconf = subtractDict(profconf, base)
			helps.select('Profile: "%s"' % prof).add(
				pformat(profconf, indent = 2, width = 100))
		config.update(default)

	print('\n'.join(HELP_ASSEMBLER.assemble(helps)), end = '')

def main():
	"""Main entry point"""
	pipelines = Pipeline.pipelines()
	command = sys.argv[1] if len(sys.argv) > 1 else None
	if command == 'params':
		_, opts, _ = commands._parse(dict_wrapper = Box)
		showParams(opts)
	elif command == 'list':
		_, opts, _ = commands._parse(dict_wrapper = Box)
		listProcs(opts)
	elif command == 'proc':
		_, opts, _ = commands._parse(dict_wrapper = Box)
		proc(opts)
	elif command in ('completion', 'completions'):
		_, opts, _ = commands._parse(dict_wrapper = Box)
		complete(opts)
	elif command == 'profile':
		_, opts, _ = commands._parse(dict_wrapper = Box)
		profile(opts)
	elif command in pipelines:
		# let the pipeline parse the arguments
		Pipeline(command).run()
	elif command in commands._hcmd:
		_, opts, _ = commands._parse(arbi = True, dict_wrapper = Box)
		if not opts._:
			commands._help(print_and_exit = True)
		if opts._ in commands._cmds:
			commands._cmds[opts._]._help(print_and_exit = True)
		if opts._ in pipelines:
			Pipeline(opts._).module.params._help(print_and_exit = True)
		if '.' in opts._: # process
			module, prc = opts._.split('.')
			try:
				Module(module).procs()[prc].printHelps()
			except KeyError:
				raise KeyError('Module %r does not have process: %s' % (module, prc)) from None
		else: # assume module
			listProcs(Box(_ = [opts._]))
	elif not sys.argv[0]:
		raise RuntimeError('This package has to run as a command line tool.')
	elif not command:
		commands._help(print_and_exit = True)
	else:
		# Hold helps first, because we haven't assembled the help page yet
		hopts = commands[command]._hopts
		commands[command]._hopts = []
		command, opts, _ = commands._parse(arbi = True, dict_wrapper = Box)
		# resume it
		commands[command]._hopts = hopts
		if '.' in command:
			module, prc = command.split('.')
			module = Module(module)
			if prc not in module.procs():
				raise ValueError('No such process: %s' % command)
			module.procs()[prc].run(opts)
		else: # listing module processes
			listProcs(Box(_ = [command]))




from os import path
from pyppl import Box
from gff import Gff
from bioprocs.utils.tsvio2 import TsvWriter, TsvReader, TsvRecord
from bioprocs.utils import logger

infile   = {{ i.infile | quote}}
outfile  = {{ o.outfile | quote}}
notfound = {{ args.notfound | quote}}
genecol  = {{ args.genecol or 0 | repr}}
inopts   = {{ args.inopts | repr}}
refgene  = {{ args.refgene | quote}}

if not path.isfile(refgene):
	raise OSError('Refgene file does not exists: {}'.format(refgene))

# get genes
genes  = TsvReader(infile, **inopts).dump(genecol)
genes  = dict(zip(genes, [False] * len(genes)))
writer = TsvWriter(outfile)
writer.cnames = ['CHR', 'START', 'END', 'NAME', 'SCORE', 'STRAND']

gff    = Gff(refgene)
for g in gff:
	attrs = g['attributes']
	if attrs['gene_id'] not in genes:
		continue
	r        = TsvRecord()
	r.CHR    = g['seqid']
	r.START  = g['start']
	r.END    = g['end']
	r.SCORE  = g['score']
	r.STRAND = g['strand']
	r.NAME   = attrs['gene_id']
	writer.write(r)
writer.close()

for g, v in genes.items():
	if not v:
		logger.warning('Gene: {!r} not found.'.format(g))



import sys
from pyppl import Box
from bioprocs.utils import shell2 as shell

bedtools = {{args.bedtools | quote}}
params   = {{ args.params | repr}}

params.a           = {{i.afile | quote}}
params.b           = {{i.bfiles | repr}}
params.wao         = params.get('wao', True)
params.nonamecheck = params.get('nonamecheck', True)
params._out        = {{o.outfile | quote}}
params._stderr     = sys.stderr

shell.load_config(bedtools = bedtools)

shell.bedtools.intersect(**params)



"""Some commonly-used processes"""
from pyppl import Proc, Box
from modkit import Modkit
from . import params, delefactory, procfactory
from .utils import fs2name
Modkit().delegate(delefactory())

@procfactory
def _pSort():
	"""
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`inopts`: The input options for infile:
			- `skip`   : First N lines to skip. Default: `0`
			- `delimit`: The delimit. Default          : `\t`
		`case`:   Case-sensitivity. Default: True
			- If True, will set $LANG as C
			- Otherwise, $LANG will be set as en_US.UTF-8
		`mem`    : The buffer size. Default: 4G
		`tmpdir` : The tmpdir.
		`unique` : Just keep the unique lines. Default: False
		`delimit`: The delimit to separate the fields. Default: '\t'
		`params` : The arguments used by `sort`
	"""
	return Box(
		desc   = 'Sort file using linux command `sort`',
		input  = "infile:file",
		output = "outfile:file:{{i.infile | bn}}",
		lang   = params.python.value,
		args   = Box(
			params = Box(),
			inopts = Box(skip = 0, delimit = '\t'),
			case   = True,
			mem    = params.mem4G.value,
			tmpdir = params.tmpdir.value,
			unique = False,
		)
	)

@procfactory
def _pShell():
	"""
	@name:
		pShell
	@description:
		Run shell command directly
	@input:
		args: A dict of parameters of the command
	@output:
		outfile: The output file
	@args:
		cmd (str): The executable.
	"""
	pShell          = Proc(desc = 'Run shell command directly')
	pShell.input    = 'args'
	pShell.output   = 'outfile:stdout:{{args.cmd}}.{{proc.suffix}}.{{job.index + 1}}.txt'
	pShell.args.cmd = None
	pShell.lang     = params.python.value
	pShell.script   = "file:scripts/common/pShell.py"
	return pShell

@procfactory
def _pFiles2Dir():
	"""
	@name:
		pFiles2Dir
	@description:
		A helper process to convert a list of files into a directory, so that some processes can take it as input
	@input:
		`infiles:files`: The input files
	@output:
		`outdir:dir`:    The output directory
	"""
	pFiles2Dir                = Proc(desc = 'Put files to a directory using symbolic links.')
	pFiles2Dir.input          = "infiles:files"
	pFiles2Dir.output         = "outdir:dir:{{i.infiles | lambda x: sorted(x) | [0] | fn}}.dir"
	pFiles2Dir.lang           = params.python.value
	pFiles2Dir.script         = "file:scripts/common/pFiles2Dir.py"
	return pFiles2Dir

@procfactory
def _pFile2Proc():
	"""
	@name:
		pFile2Proc
	@description:
		Convert a file to a proc so it can be used as dependent
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	"""
	pFile2Proc                = Proc(desc="Convert a file to a proc so it can be used as dependent")
	pFile2Proc.input          = "infile:file"
	pFile2Proc.output         = "outfile:file:{{i.infile | bn}}"
	pFile2Proc.script         = 'ln -s "{{i.infile}}" "{{o.outfile}}"'
	return pFile2Proc

@procfactory
def _pStr2File():
	"""
	@name:
		pStr2File
	@description:
		Save string to a file.
	@input:
		`in:var`: The input string.
	@output:
		`outfile:file`: The output file.
	"""
	pStr2File                 = Proc(desc = "Save string to a file.")
	pStr2File.input           = "instr:var"
	pStr2File.output          = "outfile:file:{{i.instr | encode}}.txt"
	pStr2File.args.breakOn    = ','
	pStr2File.args.trimLine   = True
	pStr2File.envs.encode     = lambda x: __import__('re').sub(r'[^\w_]', '', x)[:16]
	pStr2File.lang            = params.python.value
	pStr2File.script          = "file:scripts/common/pStr2File.py"
	return pStr2File

@procfactory
def _pHead():
	"""
	@name:
		pHead
	@description:
		Get the top N lines from a file
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`n`   : Top n lines. You may use '-n' to skip last n lines.
		`pipe`: other piped command to modify the results
	"""
	pHead           = Proc(desc = "Like linux's head command")
	pHead.input     = "infile:file"
	pHead.output    = "outfile:file:{{i.infile | fn}}.head.txt"
	pHead.args.n    = 10
	pHead.args.pipe = ''
	pHead.script    = 'head -n {{args.n}} {{i.infile | squote}} {{"| " + args.pipe if args.pipe else ""}} > {{o.outfile | squote}}'
	return pHead

@procfactory
def _pTail():
	"""
	@name:
		pTail
	@description:
		Get the bottom N lines from a file
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`n`: Bottom n lines. You may use '+n' to skip first n lines.
	"""
	pTail                     = Proc(desc = "Like linux's tail command")
	pTail.input               = "infile:file"
	pTail.output              = "outfile:file:{{i.infile | fn}}.tail.txt"
	pTail.args.n              = 10
	pTail.script              = 'tail -n {{args.n | lambda x: "+"+str(int(x)+1) if x.startswith("+") else x}} {{i.infile | squote}} > {{out.outfile | squote}}'
	return pTail

@procfactory
def _pPrepend():
	"""
	@name:
		pPrepend
	@description:
		Prepend a string to a file
	@input:
		`in:var`: The input string.
		`infile:file`: The input file.
	@output:
		`outfile:file`: The output file.
	"""
	pPrepend                  = Proc(desc = "Prepend string to a file.")
	pPrepend.input            = "in:var, infile:file"
	pPrepend.output           = "outfile:file:{{i.infile | fn2}}.prepend{{i.infile | ext}}"
	pPrepend.script           = '''
	printf {{i.in | squote}} > {{out.outfile | squote}}
	cat {{i.infile | squote}} >> {{out.outfile | squote}}
	'''
	return pPrepend

pAppend                   = Proc(desc = "Append string to a file")
pAppend.input             = "in:var, infile:file"
pAppend.output            = "outfile:file:{{i.infile | fn2}}.append{{i.infile | ext}}"
pAppend.script            = 'cat {{i.infile | squote}} > {{out.outfile | squote}}; printf {{i.in | squote}} >> {{out.outfile | squote}}'

@procfactory
def _pUnique():
	"""
	@name:
		pUnique
	@description:
		Make the input file with unique rows (at certain column)
	@input:
		`infile:file`: The input file.
	@output:
		`outfile:file`: The output file.
	@args:
		`inopts`: The options for input file
			- `delimit`: delimit for columns
			- `skip`: skip first lines
			- `comment`: signs for treating lines as comments
		`outopts`: The output options
			- `head`: Output head or not. Default: `False`
			- `headPrefix`: The prefix for the head
			- `headDelimit`: The delimit for the head
			- `headTransform`: The transform function for the head
			- `delimit`: The delimit for the data.
		`col`: The column to compare. Default: `*` (all columns)
		`sorted`: Whether the input file is sorted. Default: `False`
	"""
	pUnique                   = Proc(desc = "Make the input file unique")
	pUnique.input             = "infile:file"
	pUnique.output            = "outfile:file:{{i.infile | fn2}}.unique{{i.infile | ext}}"
	pUnique.args.inopts       = Box(delimit = "\t", skip = 0, comment = "#")
	pUnique.args.outopts      = Box(head = False, headPrefix = '', headDelimit = '\t', headTransform = None, delimit = '\t')
	pUnique.args.col          = '*'
	pUnique.args.sorted       = False
	pUnique.lang              = params.python.value
	pUnique.script            = "file:scripts/common/pUnique.py"
	return pUnique

@procfactory
def _pAddHeader():
	"""
	@name:
		pAddHeader
	@description:
		Add the header of 1st file to 2nd file.
	@input:
		`infile1:file`: The first file containing the header.
		`infile2:file`: The second file with the body.
	@output:
		`outfile:file`: The output file with the header from 1st input file, body from 2nd file.
	@args:
		`n`: The number of header lines.
	"""
	pAddHeader                = Proc(desc = 'Add the header of 1st file to 2nd file.')
	pAddHeader.input          = "infile1:file, infile2:file"
	pAddHeader.output         = "outfile:file:{{i.infile2 | bn}}"
	pAddHeader.args.n         = 1
	pAddHeader.script         = '''
	head -n {{args.n}} {{i.infile1 | squote}} > {{out.outfile | squote}}
	cat {{i.infile2 | squote}} >> {{out.outfile | squote}}
	'''
	return pAddHeader

@procfactory
def _pMergeFiles():
	"""
	@name:
		pMergeFiles
	@description:
		Merge files.
	@input:
		`infiles:files`: The input files
	@output:
		`outfile:file`: The output file
	@args:
		`header`: Whether the input files have header. Default: `False`
			- If `True`, input files must have the same header line.
	"""
	pMergeFiles              = Proc(desc = 'Merge files.')
	pMergeFiles.input        = "infiles:files"
	pMergeFiles.output       = "outfile:file:{{i.infiles | fs2name}}{{i.infiles | :a[0] if a else '' | ext }}"
	pMergeFiles.args.header  = False
	pMergeFiles.envs.fs2name = fs2name
	pMergeFiles.lang         = params.python.value
	pMergeFiles.script       = "file:scripts/common/pMergeFiles.py"
	return pMergeFiles

@procfactory
def _pGrep():
	"""
	@name:
		pGrep
	"""
	pGrep              = Proc(desc = 'Filter a file using linux grep')
	pGrep.input        = 'infile:file'
	pGrep.output       = 'outfile:file:{{i.infile | bn}}'
	pGrep.args.params  = Box()
	pGrep.args.keyword = ''
	pGrep.lang         = params.python.value
	pGrep.script       = "file:scripts/common/pGrep.py"
	return pGrep

@procfactory
def _pSplitRows():
	"""
	@name:
		pSplitRows
	@description:
		Split a file by rows, specially usefull to split a job into multithreads/multiprocesses.
	@input:
		`infile:file`: The input file
	@output:
		`outdir:dir`: The output directory including the split files
	@args:
		`skip`: The skip first n lines. Default: `0`
		`cnames`: The column names. If True, the column names will be added to each split file. Default: `True`
		`n`: Number of files to split. Default: `8`
	"""
	pSplitRows                = Proc(desc = 'Split a file by rows.')
	pSplitRows.input          = 'infile:file'
	pSplitRows.output         = 'outdir:dir:{{i.infile | bn}}.rows'
	pSplitRows.args.skip      = 0
	pSplitRows.args.cnames    = True
	pSplitRows.args.n         = 8
	pSplitRows.lang           = params.python.value
	pSplitRows.script         = "file:scripts/common/pSplitRows.py"
	return pSplitRows




"""TSV file operations"""

from pyppl import Proc, Box
from . import params, rimport
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pMatrixR():
	"""
	@name:
		pMatrixR
	@description:
		Operate a matrix and save the new matrix to file.
	@input:
		`infile:file`: The input file containing the matrix
	@output:
		`outfile:file`: The output matrix
	@args:
		`inopts`: The input options for infile:
			- `cnames`: Whether the input file has cnames. Default: True
			- `rnames  `: Whether the input file has rnames. Default: True
			- `delimit`: The delimit. Default: `\t`
			- `skip`: First N lines to skip. Default: `0`
		`params`: Other params for `read.table`. Default: `{"check.names": "FALSE", "quote": ""}`
		`code`: The R code to operating the matrix. (the matrix is read in variable `mat`)
	"""
	return Box(
		desc   = 'Operate a matrix and save the new matrix to file',
		lang   = params.Rscript.value,
		input  = "infile:file",
		output = "outfile:file:{{i.infile | bn}}",
		args   = Box(inopts = Box(cnames = True, rnames = True, delimit = "\t", skip = 0),
			params = Box({
				"check.names": "FALSE",
				"quote"      : ""
			}),
			code = []))

@procfactory
def _pTranspose():
	"""
	@name:
		pTranspose
	@description:
		Transpose a matrix
	@input:
		`infile:file`: The input matrix file
	@output:
		`outfile:file`: The Transposed matrix file. Default: `{{i.infile | bn}}`
	@args:
		`inopts`: Input options for input file.
	"""
	pTranspose              = Proc(desc = 'Transpose a matrix')
	pTranspose.input        = 'infile:file'
	pTranspose.output       = 'outfile:file:{{i.infile | bn}}'
	pTranspose.args.inopts  = Box(cnames = True, rnames = True)
	pTranspose.envs.rimport = rimport
	pTranspose.lang         = params.Rscript.value
	pTranspose.script       = "file:scripts/tsv/pTranspose.r"
	return pTranspose

@procfactory
def _pPaired():
	"""
	@name:
		pPaired
	@description:
		Subset each input file and make sure they have paired columns.
	@input:
		`infile1:file`: The first file
		`infile2:file`: The second file
	@outfile:
		`outfile1:file`: The paired file for infile1. Default: `{{i.infile1 | fn}}.paired{{i.infile1 | ext}}`
		`outfile2:file`: The paired file for infile2. Default: `{{i.infile2 | fn}}.paired{{i.infile2 | ext}}`
	@args:
		`inopts1`: reading options for input file1
		`inopts2`: reading options for input file2
	"""
	pPaired        = Proc(desc = "Subset each input file and make sure they have paired columns.")
	pPaired.input  = 'infile1:file, infile2:file'
	pPaired.output = [
		'outfile1:file:{{i.infile1 | fn}}.paired{{i.infile1 | ext}}',
		'outfile2:file:{{i.infile2 | fn}}.paired{{i.infile2 | ext}}'
	]
	pPaired.args.inopts1 = Box(head = True, headCallback = None)
	pPaired.args.inopts2 = Box(head = True, headCallback = None)
	pPaired.lang         = params.python.value
	pPaired.script       = "file:scripts/tsv/pPaired.py"
	return pPaired


@procfactory
def _pCbind():
	"""
	@name:
		pCbind
	@description:
		Cbind the rest of files to the first file.
	@input:
		`infiles:files`: The input files
	@output:
		`outfile:file`: The output matrix
	@args:
		`inopts`: The input options for infile:
			- `cnames`: Whether the input file has cnames. Default: True
			- or [True, True, False] corresponding to the file order
			- `rnames  `: Whether the input file has rnames. Default: True
			- `delimit`: The delimit. Default: `\t`
			- `skip`: First N lines to skip. Default: `0`
		`params`: Other params for `read.table`. Default: `{"check.names": "FALSE", "quote": ""}`
		`fn2cname`: The function (r) used to convert file name to column name.
		`fill`: Do `cbind.fill` instead of `cbind`. Default: `True`
			- Set it to `False` if the row names are in the same order
		`na`: Replacement for missing values. Default: `NA`
	"""
	pCbind             = Proc(desc = 'Cbind the rest of files to the first file.')
	pCbind.input       = 'infiles:files'
	pCbind.output      = 'outfile:file:{{i.infiles | fs2name}}.cbound.txt'
	pCbind.args.inopts = Box(
		cnames  = True, # or [True, True, False] corresponding to the file order
		rnames  = True,
		delimit = "\t",
		skip    = 0
	)
	pCbind.args.params = Box({
		"check.names": "FALSE",
		"quote"      : ""
	})
	pCbind.args.na       = 'NA'
	pCbind.args.fn2cname = 'function(fn) fn' # if
	pCbind.args.fill     = True
	pCbind.envs.fs2name  = fs2name
	pCbind.envs.rimport  = rimport
	pCbind.lang          = params.Rscript.value
	pCbind.script        = "file:scripts/tsv/pCbind.r"
	return pCbind

@procfactory
def _pRbind():
	"""
	@name:
		pRbind
	@description:
		Rbind the rest of files to the first file.
	@input:
		`infiles:files`: The input files
	@output:
		`outfile:file`: The output matrix
	@args:
		`inopts`: The input options for infile:
			- `cnames`: Whether the input file has cnames. Default: True
			- or [True, True, False] corresponding to the file order
			- `rnames  `: Whether the input file has rnames. Default: True
			- `delimit`: The delimit. Default: `\t`
			- `skip`: First N lines to skip. Default: `0`
		`params`: Other params for `read.table`. Default: `{"check.names": "FALSE", "quote": ""}`
		`na`: Replacement for missing values. Default: `NA`
		`fn2rname`: The function (r) used to convert file name to row name.
		`fill`: Do `rbind.fill` instead of `rbind`. Default: `True`
			- Set it to `False` if the row names are in the same order
		`na`: Replacement for missing values. Default: `NA`
	"""
	pRbind                = Proc(desc = 'Rbind the rest of files to the first file.')
	pRbind.input          = 'infiles:files'
	pRbind.output         = 'outfile:file:{{i.infiles[0] | bn}}'
	pRbind.args.inopts    = Box(
		cnames  = True, # or [True, True, False] corresponding to the file order
		rnames  = True,
		delimit = "\t",
		skip    = 0
	)
	pRbind.args.params = Box({
		"check.names": "FALSE",
		"quote"      : ""
	})
	pRbind.args.na       = 'NA'
	pRbind.args.fn2rname = 'function(fn) fn'
	pRbind.args.fill     = True
	pRbind.envs.fs2name  = fs2name
	pRbind.envs.rimport  = rimport
	pRbind.lang          = params.Rscript.value
	pRbind.script        = "file:scripts/tsv/pRbind.r"
	return pRbind

@procfactory
def _pCsplit():
	"""
	@name:
		pCsplit
	@description:
		Split a matrix by columns and save them into files.
	@input:
		`infile:file`: The input file
	@output:
		`outdir:dir`: The directory containing the output column files
	@args:
		`inopts`: The input options for infile:
			- `cnames`: Whether the input file has cnames. Default: True
			- `rnames  `: Whether the input file has rnames. Default: True
			- `delimit`: The delimit. Default: `\t`
			- `skip`: First N lines to skip. Default: `0`
		`params`: Other params for `read.table`. Default: `{"check.names": "FALSE", "quote": ""}`
		`size`: The chunk size (how many columns to split into one file). Default: `1`
	"""
	pCsplit             = Proc(desc = 'Split the columns of input file into different files.')
	pCsplit.input       = 'infile:file'
	pCsplit.output      = 'outdir:dir:{{i.infile | fn}}.csplits'
	pCsplit.args.inopts = Box(
		cnames  = True,
		rnames  = True,
		delimit = "\t",
		skip    = 0
	)
	pCsplit.args.params = Box({
		"check.names": "FALSE",
		"quote"      : ""
	})
	pCsplit.args.size    = 1
	pCsplit.envs.rimport = rimport
	pCsplit.lang         = params.Rscript.value
	pCsplit.script       = "file:scripts/tsv/pCsplit.r"
	return pCsplit

@procfactory
def _pRsplit():
	"""
	@name:
		pRsplit
	@description:
		Split a matrix by rows and save them into files.
	@input:
		`infile:file`: The input file
	@output:
		`outdir:dir`: The directory containing the output row files
	@args:
		`inopts`: The input options for infile:
			- `cnames`: Whether the input file has cnames. Default: True
			- `rnames  `: Whether the input file has rnames. Default: True
			- `delimit`: The delimit. Default: `\t`
			- `skip`: First N lines to skip. Default: `0`
		`params`: Other params for `read.table`. Default: `{"check.names": "FALSE", "quote": ""}`
		`size`: The chunk size (how many rows to split into one file). Default: `1`
	"""
	pRsplit             = Proc(desc = 'Rbind the rest of files to the first file.')
	pRsplit.input       = 'infile:file'
	pRsplit.output      = 'outdir:dir:{{i.infile | fn}}.rsplits'
	pRsplit.args.inopts = Box(
		cnames  = True,
		rnames  = True,
		delimit = "\t",
		skip    = 0
	)
	pRsplit.args.params = Box({
		"check.names": "FALSE",
		"quote"      : ""
	})
	pRsplit.args.size    = 1
	pRsplit.envs.rimport = rimport
	pRsplit.lang         = params.Rscript.value
	pRsplit.script       = "file:scripts/tsv/pRsplit.r"
	return pRsplit

@procfactory
def _pTsv():
	"""
	@name:
		pTsv
	@description:
		Read, Transform, filter a TSV file.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`inopts`: The input options for infile:
		`outopts`: The output options for outfile
		`row`: A row function to transform/filter the row. Argument is an instance of `TsvRecord`
		`helper`: A helper function for `args.ops`
	"""
	pTsv              = Proc(desc = 'Read, Transform, filter a TSV file.')
	pTsv.input        = "infile:file"
	pTsv.output       = "outfile:file:{{i.infile | fn}}.tsv"
	pTsv.lang         = params.python.value
	pTsv.args.helper  = ''
	pTsv.args.row     = None
	pTsv.args.inopts  = Box(delimit = '\t', comment = '#', skip = 0, cnames = True)
	pTsv.args.outopts = Box(delimit = '\t', cnames = True)
	pTsv.script       = "file:scripts/tsv/pTsv.py"
	return pTsv

@procfactory
def _pTsvColFilter(alias = 'pTsvColSelect'):
	"""
	@input:
		infile : The input file
		colfile: The file with columns, one per line, or a list of columns separated by comma.
			- If this is provided, `args.cols` will be ignored.
	@output:
		outfile: The output file
	@args:
		inopts: The options for reading input file. Default: `Box(cnames = True)`
		keep  : Whether to keep in `args.cols` or to discard
		cols  : The columns used to filter. Could be names or indices(0-based) or a file containing the column names, one per line.
	"""
	return Box(
		desc   = 'Filter a tsv file by columns',
		lang   = params.python.value,
		input  = 'infile:file, colfile:var',
		output = 'outfile:file:{{i.infile | bn}}',
		args   = Box(
			inopts = Box(cnames = True),
			keep = True,
			cols = None,
		)
	)

@procfactory
def _pTsvAggregate():
	"""
	@name:
		pTsvAggregate
	@description:
		Aggregate on columns with a set of records
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn2}}.aggr.txt`
			- With columns `args.on` and aggregated results from `args.aggrs`
			- If `args.on` is a function, then the calculated term will be add to the 1st column.
	@args:
		`inopts`: The options to read the input file, Default: `Box(cnames = True)`
		`on`: Aggregate according to which column, Default: `0`
			- It also could column name if `args.inopts = True`
			- The input file has to sorted by this column
			- Or a string of (lambda) function to calculate the term to aggregate on.
		`helper`: Raw codes to give some help for `args.aggrs`
		`aggrs`: The aggregation methods. Required.
			- It's a `Box` with the keys for aggregated results
			- If `args.inopts.cnames = True` then these keys will be output as column names, otherwise ignored
			- You can also combine the aggregation results.
				- For example: `{"sum,mean": lambda rs: [sum(r.value for r in rs), sum(r.value for r in rs)/float(len(rs))]}`
			- We have some built-in aggregation methods:
				- `args.aggrs.Sum = "$sum:3"`          : Get the sum of the 4th column
				- `args.aggrs.Mean = "$mean:Height`    : Get the mean of column "Height"
				- `args.aggrs.Median = "$median:1"`    : Get the median of the 2nd column
				- `args.aggrs.Min = "$min:1"`          : Get the min of the 2nd column
				- `args.aggrs.Max = "$max:1"`          : Get the max of the 2nd column
				- `args.aggrs.Max2 = "$max:2"`         : Get the max of the 3rd column
				- `args.aggrs["Max,Max2"] = "$max:1,2"`: Get the max of the 2nd and 3rd column, respectively
				- `args.aggrs.CombinedP = "$fisher:1"` : Get the combined pvalues for 1st column using fisher'method (`scipy.stats.combine_pvalues`)
	"""
	pTsvAggregate             = Proc(desc = 'Aggregate on columns with a set of records')
	pTsvAggregate.input       = 'infile:file'
	pTsvAggregate.output      = 'outfile:file:{{i.infile | fn2}}.aggr.txt'
	pTsvAggregate.args.inopts = Box(cnames = True)
	pTsvAggregate.args.on     = 0 # which column
	pTsvAggregate.args.aggrs  = Box()
	pTsvAggregate.args.helper = ''
	pTsvAggregate.lang        = params.python.value
	pTsvAggregate.script      = "file:scripts/tsv/pTsvAggregate.py"
	return pTsvAggregate

@procfactory
def _pTsvHeader():
	"""
	@name:
		pTsvHeader
	@description:
		Get the header of a TSV file
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn2}}.header.txt`
	@args:
		`inopts`: The options to read input file. Default: `Box(cnames = True)`
		`filter`: The filter for the header. Default: `None`
			- `None`: no filter
			- `lambda cnames: ...` A callback to manipulate colnames.
	"""
	pTsvHeader             = Proc(desc = 'Get the header of a tsv file.')
	pTsvHeader.input       = 'infile:file'
	pTsvHeader.output      = 'outfile:file:{{i.infile | fn2}}.header.txt'
	pTsvHeader.args.inopts = Box(cnames = True)
	pTsvHeader.args.filter = None
	pTsvHeader.lang        = params.python.value
	pTsvHeader.script      = "file:scripts/tsv/pTsvHeader.py"
	return pTsvHeader

@procfactory
def _pTsvReplaceHeader():
	"""
	@name:
		pTsvReplaceHeader
	@description:
		Replace the header of a TSV file
	@input:
		`infile:file`: The input file
		`hfile:file`:  The file containing the headers, one per line.
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | bn}}`
	@args:
		`inopts`: The options to read input file, Default: `Box(cnames = True)`
		`cnames`: The column names or callback, Default: `None`
			- `None`: use the header in `i.hfile`
			- `<list/str/file>`: the header to use if `i.hfile` is not provided
			- `lambda cnames: ...`: The callback to modify header in `i.hfile` if provided, otherwise modify the original header.
	"""
	pTsvReplaceHeader             = Proc(desc = "Replace the header of a tsv file.")
	pTsvReplaceHeader.input       = 'infile:file, hfile:file'
	pTsvReplaceHeader.output      = 'outfile:file:{{i.infile | bn}}'
	pTsvReplaceHeader.args.inopts = Box(cnames = True)
	pTsvReplaceHeader.args.cnames = None
	pTsvReplaceHeader.lang        = params.python.value
	pTsvReplaceHeader.script      = "file:scripts/tsv/pTsvReplaceHeader.py"
	return pTsvReplaceHeader

@procfactory
def _pTsvJoin():
	"""
	@name:
		pTsvJoin
	@description:
		Read files simultaneously.
		NOTE: only one file allows multiple lines with same value to compare, and that file should be the first one. For example:
		```
		File1:
		1	1
		1	2
		1	3
		File2:
		1	1
		2	2
		3	3
		```
		If you compare the first column, File1 has to put at the begining for input.
	@input:
		infiles:files: The input files
	@output:
		outfile:file: The output file
	@args:
		inopts: The input options for infile, each can be a list if different for `infiles`:
			- skip   : First N lines to skip. Default: `0`
			- delimit: The delimit. Default: `\t`
			- comment: The comment line mark. Default: `#`
			- cnames : Whether input file has head. Default: `True`
		outopts: The output options:
			- delimit: The delimit. Default: `\t`
			- cnames : Whether to output the head? Default: `False`
			- headCallback: a string callback to organize headers.
				- E.g: `lambda cnames: ','.join(cnames)`
				- The argument `cnames` will be the joint cnames from all readers
				- Could be `True` to use `\t` to assemble the headers
		debug: Save debug information in stderr file. Default: `False`
		match: The match function.
			- Return -1 if matched, otherwise return index the of record that should be read next.
			- E.g: `lambda r1, r2: -1 if r1.id == r2.id else 0 if r1.id < r2.id else 1`
			- By default, will use the first column to compare
		do: The do function. Global vaiable `fout` is available to write results to output file.
			- E.g: `lambda writer, r1, r2: writer.write(r1)`
		helper: Some helper codes.
	"""
	pTsvJoin              = Proc(desc = 'Read files simultaneously.')
	pTsvJoin.input        = 'infiles:files'
	pTsvJoin.output       = 'outfile:file:{{i.infiles[0] | fn}}.etc.joined.txt'
	pTsvJoin.echo         = 0
	pTsvJoin.args.inopts  = Box(delimit = '\t', skip = 0, comment = '#', cnames = True)
	pTsvJoin.args.outopts = Box(delimit = '\t', cnames = False)
	pTsvJoin.args.debug   = False
	pTsvJoin.args.match   = None
	pTsvJoin.args.do      = None
	pTsvJoin.args.helper  = ''
	pTsvJoin.lang         = params.python.value
	pTsvJoin.script       = "file:scripts/tsv/pTsvJoin.py"
	return pTsvJoin

@procfactory
def _pTsvSql():
	"""
	@name:
		pTsvSql
	@description:
		Query tsv file using SQL. (see: http://harelba.github.io/q/examples.html)
	@input:
		`infile:file` : The input tsv file
		`sqlfile:file`: The file containing the SQLs. If provided, `args.sql` will be ignored.
	@output:
		`outfile:file`: The output file
	@args:
		`sql`: If SQL to execute. Use `-` for table name
		`inopts`: Options for input file.
			- `cnames`: Input file has header? Default: `True`
			- `delimit`: The delimit of input file. Default: `\t`
			- `encoding`: Encoding of input file. Default: `UTF-8`
			- `gz`: Whether input file is gzipped. Default: `auto` (detected from file extension)
		`outopts`: Output options.
			- `cnames`: Inherited from `args.inopts`
			- `delimit`: Inherited from `args.inopts`
			- `encoding`: Inherited from `args.inopts`
	@requires:
		[`q`](http://harelba.github.io/q/index.html)
		This process is built on `q 1.7.1`
	"""
	pTsvSql             = Proc(desc = 'Query tsv file using SQL')
	pTsvSql.input       = 'infile:file, sqlfile:file'
	pTsvSql.output      = 'outfile:file:{{i.infile | fn}}.bysql{{i.infile | ext}}'
	pTsvSql.args.sql    = ''
	pTsvSql.args.inopts = Box(
		cnames = True, delimit = "\t", encoding = 'UTF-8', gz = 'auto')
	pTsvSql.args.outopts = Box(
		cnames = None, delimit = None, encoding = None)
	pTsvSql.lang   = params.python.value
	pTsvSql.script = "file:scripts/tsv/pTsvSql.py"
	return pTsvSql

@procfactory
def _pTsvSample():
	"""
	@name:
		pTsvSample
	@description:
		Sample records from a TSV file
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn2}}.sampled.txt`
	@args:
		`inopts`   : input options, only skip available, Default: `Box()`
		`n`        : how many records to sample, Default: `10`
		`arsample` : sample program by Alex Reynolds, Default: `<params.arsample>`
		`replace`  : Whether sample with replacement or not, Default: `False`
		`keeporder`: Keep the order of the sampled records as it's in input file, Default: `False`
		`seed`: The seed, Default: `0`
		`params`: Other params for arsample, Default: `Box()`
	"""
	pTsvSample                = Proc(desc = 'Sample records from a TSV file.')
	pTsvSample.input          = 'infile:file'
	pTsvSample.output         = 'outfile:file:{{i.infile | fn2}}.sampled.txt'
	pTsvSample.args.inopts    = Box()
	pTsvSample.args.n         = 10
	pTsvSample.args.arsample  = params.arsample.value
	pTsvSample.args.replace   = False
	pTsvSample.args.keeporder = False
	pTsvSample.args.seed      = 0
	pTsvSample.args.params    = Box()
	pTsvSample.lang           = params.python.value
	pTsvSample.script         = "file:scripts/tsv/pTsvSample.py"
	return pTsvSample

@procfactory
def _pTsvMerge():
	"""
	@name:
		pTsvMerge
	@description:
		Merge files in the input directory
	@input:
		`indir:file`: The input directory
	@output:
		`outfile:file`: The output file
	@args:
		`inopts`: The options for input file. Default: `Box(skip = 0, comment = '#', delimit = '\t')`
		`outopts`: The options for output file. Default: `Box()`
	"""
	pTsvMerge               = Proc(desc = 'Merge files by rows.')
	pTsvMerge.input         = "infiles:files"
	pTsvMerge.output        = "outfile:file:{{i.infiles | fs2name}}"
	pTsvMerge.args.inopts   = Box(skip = 0, comment = '#', delimit = '\t')
	pTsvMerge.args.outopts  = Box()
	# IOError: [Errno 24] Too many open files
	pTsvMerge.args.maxopen  = 100
	pTsvMerge.envs.fs2name  = fs2name
	pTsvMerge.lang          = params.python.value
	pTsvMerge.script        = "file:scripts/tsv/pTsvMerge.py"
	return pTsvMerge

@procfactory
def _pMergeRows():
	"""
	@name:
		pMergeRows
	@description:
		Merge repeated rows
	@input:
		`infile:file`: The input file (has to be sorted by the repeated columns)
	@output:
		`outfile:file`: The output file. Default: `{{i.infile | bn}}`
	@args:
		`inopts`: The options for input file.
			- defaults: skip: 0, comment: #, delimit '\\t'
		`outopts`: The options for output file. Defaults:
			- head: False (not output head line)
			- headPrefix: `#` (The prefix for head line)
			- headDelimit: `\\t` (The delimit for head line)
			- headTransform: `None` (The callback for head line)
			- delimit: `\\t` (The delimit for output line)
		`match`: The function to return a value to decide whether the row is repeated, argument is a `TsvRecord`.
		`do`   : The merge function in python, argument is a list of `TsvRecord`s or `list`s if `args.inopts.ftype` is `nometa`
	"""
	pMergeRows              = Proc(desc = 'Merge repeated rows.')
	pMergeRows.input        = 'infile:file'
	pMergeRows.output       = 'outfile:file:{{i.infile | bn}}'
	pMergeRows.args.inopts  = Box(skip = 0, comment = '#', delimit = '\t')
	pMergeRows.args.outopts = Box(head = False, headPrefix = '', headDelimit = '\t', headTransform = None, delimit = '\t')
	pMergeRows.args.match   = None
	pMergeRows.args.do      = None
	pMergeRows.lang         = params.python.value
	pMergeRows.script       = "file:scripts/tsv/pMergeRows.py"
	return pMergeRows




from pyppl import Box

bcode   = {{args.bcode | repr}}
acode   = {{args.acode | repr}}
infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
scale   = {{args.scale | repr}}
section = {{args.section | repr}}
legend  = {{args.legend | render | repr}}
{% case True %}
{% when isinstance(args.title, (tuple, list)) %}
title = {{args.title | lambda a, render = render: (render(a[0]), a[1])}}
{% when args.title | bool %}
title = {{args.title | lambda a, render = render: (render(a), 1)}}
{% else %}
title = {{args.title | repr}}
{% endcase %}
align   = {{args.align | repr}}

# fix parameters
if section:
	section['type']   = section.get('type', 'NEW_PAGE').upper()
	section['orient'] = section.get('orient', 'PORTRAIT').upper()
	section['margin'] = section.get('margin', [36] * 4)
	if isinstance(section['margin'], int):
		section['margin'] = [section['margin']] * 4
	elif len(section['margin']) == 2:
		section['margin'].extend(section['margin'])
	elif len(section['margin']) == 3:
		section['margin'].append(section['margin'][1])

if not isinstance(bcode, (tuple, list)):
	bcode = [bcode]
bcode = "\n".join(bcode)

if not isinstance(acode, (tuple, list)):
	acode = [acode]
acode = "\n".join(acode)

scale = ', '.join('{} = Pt({})'.format(k, v) for k, v in scale.items())
scale = ', ' + scale if scale else ''

if legend:
	legend = legend.split('.', 1) if legend.startswith('Figure ') and '.' in legend else ['', legend]

# region template
template = """
# Set up section
{section_code}

# Add heading
{heading_code}

# Interpolation code before image
{bcode_code}

# Insert the image
{image_code}

# Add legend
{legend_code}

# Interpolation code after image
{acode_code}
"""
# endregion

# region section_code
section_code = """
sec             = doc.add_section()
sec.type        = WD_SECTION.{sectype}
sec.orientation = WD_ORIENT.{secorient}

_new_width, _new_height = sec.page_height, sec.page_width
if sec.orientation != doc.sections[-2].orientation:
	sec.page_width, sec.page_height = _new_width, _new_height
else:
	sec.page_width, sec.page_height = _new_height, _new_width
sec.top_margin, sec.right_margin, sec.bottom_margin, sec.left_margin = (Pt(x) for x in {secmargin!r})
""".format(
	sectype   = section['type'],
	secorient = section['orient'],
	secmargin = section['margin']
) if section else ""
# endregion

# region heading_code
heading_code = """
doc.add_heading('''{}''', {})
""".format(*title) if title else ""
# endregion

bcode_code = bcode

image_code = """
para = doc.add_paragraph()
{alignment}
run = para.add_run()
run.add_picture({pic!r}{scale})
""".format(
	alignment = "para.alignment = WD_ALIGN_PARAGRAPH.{}\n".format(align.upper()) if align else '',
	pic       = infile,
	scale     = scale
)

legend_code = """
_legend_para = doc.add_paragraph()
{alignment}
_legend_para.add_run({legend0!r}).font.bold = True
_legend_para.add_run({legend1!r})
""".format(
	alignment = "_legend_para.alignment = WD_ALIGN_PARAGRAPH.{}\n".format(align.upper()) if align else '',
	legend0   = legend[0],
	legend1   = legend[1]
) if legend else ''

acode_code = acode

with open(outfile, 'w') as f:
	f.write(template.format(
		section_code = section_code,
		heading_code = heading_code,
		bcode_code   = bcode_code,
		image_code   = image_code,
		legend_code  = legend_code,
		acode_code   = acode_code
	))




from os import makedirs, path
from pyppl import Box
from bioprocs.utils import runcmd, mem2, cmdargs

tmpdir = path.join({{args.tmpdir | quote}}, "tmp.{{proc.id}}.{{proc.tag}}.{{proc.suffix}}.{{job.index}}")
if not path.exists(tmpdir):
	makedirs(tmpdir)

infile  = {{i.infile | quote}}
fqfile  = {{o.fqfile | quote}}
{% if args.gz %}
fqfile = fqfile[:-3]
{% endif %}

params  = {{args.params}}
try:
{% case args.tool %}
	############# biobambam
	{% when 'biobambam' %}
	params['gz']       = 0
	#bug
	#params['S']        = fqfile
	params['filename'] = infile
	params['T']        = path.join(tmpdir, infile + '.tmp')
	if infile.endswith('.sam'):
		params['inputformat'] = 'sam'

	cmd = '{{args.biobambam}} %s > "%s"' % (cmdargs(params, dash = '', equal = '='), fqfile)
	runcmd (cmd)

	############# bedtools
	{% when 'bedtools' %}
	params['i']  = infile
	params['fq'] = fqfile

	cmd = '{{args.bedtools}} bamtofastq %s' % cmdargs(params, dash = '-', equal = ' ')
	runcmd (cmd)

	############# samtools
	{% when 'samtools' %}
	params['t'] = True
	params['s'] = fqfile

	cmd = '{{args.samtools}} fastq %s "%s"' % (cmdargs(params), infile)
	runcmd (cmd)

	############# picard
	{% when 'picard' %}
	mem = mem2({{ args.mem | quote }}, 'java')
	params['TMP_DIR'] = tmpdir
	params['I'] = infile
	params['F'] = fqfile
	cmd = '{{args.picard}} SamToFastq %s -Djava.io.tmpdir="%s" %s' % (mem, tmpdir, cmdargs(params, dash = '', equal = '='))
	runcmd (cmd)
{% endcase %}

	{% if args.gz %}
	runcmd ('gzip "%s"' % (fqfile))
	{% endif %}
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise
finally:
	from shutil import rmtree
	rmtree (tmpdir)



from os import path
from pyppl import Box
from shutil import copyfile
from bioprocs.utils import shell2 as shell

cnvkit   = {{args.cnvkit | quote}}
cnrfile  = {{i.cnrfile | quote}}
cnsfile  = {{i.cnsfile | quote}}
stem     = {{i.cnrfile | bn | quote}}
outdir   = {{o.outdir | quote}}
nthread  = {{args.nthread | repr}}
params   = {{args.params | repr}}

shell.load_config(cnvkit = dict(
	_exe = cnvkit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	),
	_cwd = outdir
))

# also report cnr, cns files
shell.cp(cnrfile, path.join(outdir, path.basename(cnrfile)))
shell.cp(cnsfile, path.join(outdir, path.basename(cnsfile)))

if params.breaks:
	if params.breaks is True:
		params.breaks = Box()
	params.breaks.o = path.join(outdir, stem + '.breaks.txt')
	shell.fg.cnvkit.breaks(cnrfile, cnsfile, **params.breaks)

if params.gainloss:
	if params.gainloss is True:
		params.gainloss = Box()
	params.gainloss.o = path.join(outdir, stem + '.gainloss.txt')
	params.gainloss.s = cnsfile
	shell.fg.cnvkit.gainloss(cnrfile, **params.gainloss)

if params.metrics:
	if params.metrics is True:
		params.metrics = Box()
	params.metrics.o = path.join(outdir, stem + '.metrics.txt')
	params.metrics.s = cnsfile
	shell.fg.cnvkit.metrics(cnrfile, **params.metrics)

if params.segmetrics:
	if params.segmetrics is True:
		params.segmetrics = Box()
	params.segmetrics.o = path.join(outdir, stem + '.segmetrics.txt')
	params.segmetrics.s = cnsfile
	shell.fg.cnvkit.segmetrics(cnrfile, **params.segmetrics).run()




from glob import glob
from os import path, symlink
from sys import stderr

for fname in {{i.infiles}}:
	bn  = path.basename (fname)
	dst = path.join ({{o.outdir | quote}}, bn)
	if path.exists (dst):
		if path.samefile(fname, dst):
			continue

		fn, ext = path.splitext(bn)
		dsts    = glob (path.join("{{o.outdir}}", fn + "[[]*[]]" + ext))

		if not dsts:
			dst = path.join("{{o.outdir}}", fn + "[1]" + ext)
		else:
			if any([path.samefile(fname, d) for d in dsts]):
				continue

			maxidx = max([int(path.basename(d)[len(fn)+1 : -len(ext)-1]) for d in dsts])
			dst = path.join("{{o.outdir}}", fn + "[" + str(maxidx+1) + "]" + ext)
		stderr.write ("pyppl.log.warning: rename %s to %s\n" % (bn, path.basename(dst)))
	symlink (fname, dst)



"""
Some long range interaction processes, not necessarily Hi-C
"""
from pyppl import Proc, Box
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pPartners():
	"""
	@name:
		pPartners
	@description:
		Find the interaction partners of the regions in input file.
	@input:
		`regfile:file`: The region file for the regions to find partners for.
		`intfile:file`: The interaction file
	@output:
		`outfile:file`: The regions with partners.
	@args:
		`regtype`: The type of region file. Default: `auto` (tell from file extension)
			- Could also be `bed` or `bedx`
		`inttype`: The type of interaction file. Default: `auto`
			- Could also be `bedpe`, `chiapet.tool`, `hiclib` and `bed12`
	"""
	pPartners                = Proc(desc = 'Find the interaction partners of the regions in input file.')
	pPartners.input          = "regfile:file, intfile:file"
	pPartners.output         = "outfile:file:{{i.regfile | fn2}}-{{i.intfile | fn2}}.partners.bedx"
	pPartners.args.regopts   = Box(ftype = "auto") # bed,   bedx
	pPartners.args.intopts   = Box(ftype = "auto") # bedpe, chiapet.tool, hiclib, bed12
	pPartners.lang           = params.python.value
	pPartners.script         = "file:scripts/hic/pPartners.py"
	return pPartners




"""Plot genomic features using Gviz R package"""
# https://bioconductor.org/packages/devel/bioc/vignettes/Gviz/inst/doc/Gviz.pdf
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())


@procfactory
def _pInteractionTrack():
	"""
	@name:
		pInteractionTrack
	@description:
		Gererate genomic interaction track for Gviz
	@input:
		`name`: The name of the track
		`infile:file`: The input file.
			- See the `type` argument for `makeGenomicInteractionsFromFile` from `GenomicInteractions` r-package
		`region`: the region, just chromosome!
	@output:
		`outfile:file`: The dumped track data
	@args:
		`intype`: Input file type. Default: auto
			- Identified by extension
			- One of "chiapet.tool", "bed12", "bedpe", "hiclib", "homer", "bam", "two.bams".
		`params`: The display params
	"""
	pInteractionTrack             = Proc(desc = 'Gererate genomic interaction track for Gviz.')
	pInteractionTrack.input       = "name, infile:file, region"
	pInteractionTrack.output      = "outfile:file:interactionTrack_{{i.name}}_{{i.region | lambda x: x.replace(':', '-')}}.gviztrack"
	pInteractionTrack.args.intype = "auto"
	pInteractionTrack.args.params = Box({
		'background.title': "#333333",
		'col'             : 'NULL',
		'col.outside'     : "#99beff",
		'anchor.height'   : 0.06,
		'col.anchors.fill': '#99beff',
		'col.anchors.line': '#AAAAAA',
		'col.interactions': '#f98d85',
		'size'            : 2
	})
	pInteractionTrack.lang        = params.Rscript.value
	pInteractionTrack.script      = "file:scripts/genomeplot/pInteractionTrack.r"
	return pInteractionTrack

@procfactory
def _pGeneTrack():
	"""
	@name:
		pGeneTrack
	@description:
		Generate gene track using ucsc data source
	@input:
		`name`:   The name of the track
	@output:
		`outfile:file`: The file to save the track
	@args:
		`genome`: The genome
		`params`: use `displayPars(UcscTrack(genome="mm9", chromosome="chrM", track="knownGene"))` to see all available args
	@requires:
		[r-Gviz](https://rdrr.io/bioc/Gviz)
	"""
	pGeneTrack             = Proc(desc = 'Generate gene track data for pGenomePlot.')
	pGeneTrack.input       = "name, region"
	pGeneTrack.output      = "outfile:file:geneTrack_{{i.name}}_{{i.region | lambda x: x.replace(':', '-')}}.gviztrack"
	pGeneTrack.args.genome = params.genome.value
	pGeneTrack.args.params = Box()
	pGeneTrack.lang        = params.Rscript.value
	pGeneTrack.script      = "file:scripts/genomeplot/pGeneTrack.r"
	return pGeneTrack

@procfactory
def _pAnnoTrack():
	"""
	@name:
		pAnnoTrack
	@description:
		The annotation track of Gviz
	@input:
		`name`:        the name of the track
		`infile:file`: the file for the track. (wig, bigWig or bedGraph, bam, need to be indexed!)
		`chrom`:       the chrom
	@output:
		`outfile:file`:the rds file for the track
	@args:
		`genome`: The genome
		`params`:  See `displayPars(DataTrack())` for all available display params
	@requires:
		[r-Gviz](https://rdrr.io/bioc/Gviz/man/DataTrack-class.html)
	"""
	pAnnoTrack             = Proc(desc = 'Generate annotation track for pGenomePlot.')
	pAnnoTrack.input       = "name, infile:file, chrom"
	pAnnoTrack.output      = "outfile:file:dataTrack_{{i.name}}_{{i.chrom}}.gviztrack"
	pAnnoTrack.args.genome = params.genome.value
	pAnnoTrack.args.params = Box()
	pAnnoTrack.lang        = params.Rscript.value
	pAnnoTrack.script      = "file:scripts/genomeplot/pAnnoTrack.r"
	return pAnnoTrack

@procfactory
def _pDataTrack():
	"""
	@name:
		pDataTrack
	@description:
		The data track of Gviz
	@input:
		`name`:        the name of the track
		`infile:file`: the file for the track. (wig, bigWig or bedGraph, bam, need to be indexed!)
		`chrom`:       the chrom
	@output:
		`outfile:file`:the rds file for the track
	@args:
		`genome`: The genome
		`params`:  See `displayPars(DataTrack())` for all available display params
	@requires:
		[r-Gviz](https://rdrr.io/bioc/Gviz/man/DataTrack-class.html)
	"""
	pDataTrack             = Proc(desc = 'Generate data track for pGenomePlot.')
	pDataTrack.input       = "name, infile:file, chrom"
	pDataTrack.output      = "outfile:file:dataTrack_{{i.name}}_{{i.chrom}}.gviztrack"
	pDataTrack.args.genome = params.genome.value
	pDataTrack.args.params = Box()
	pDataTrack.lang        = params.Rscript.value
	pDataTrack.script      = "file:scripts/genomeplot/pDataTrack.r"
	return pDataTrack

@procfactory
def _pUcscTrack():
	"""
	@name:
		pUcscTrack
	@description:
		Generate track from ucsc
	@input:
		`name`     : the name of the track
		`track`    : the UCSC track
		`trackType`: the Gviz track
		`region`   : the region
	@output:
		`outfile:file`:the dumped track
	@args:
		`params`: use `displayPars(UcscTrack(genome="mm9", chromosome="chrM", track="knownGene"))` to see all available args.
	@requires:
		[r-Gviz](https://rdrr.io/bioc/Gviz)
	"""
	pUcscTrack             = Proc(desc = "Generate track from UCSC data.")
	pUcscTrack.input       = "name, track, trackType, region"
	pUcscTrack.output      = "outfile:file:ucscTrack_{{i.name}}_{{i.region | lambda x: x.replace(':', '-')}}.gviztrack"
	pUcscTrack.lang        = params.Rscript.value
	pUcscTrack.args.genome = params.genome.value
	pUcscTrack.args.params = Box()
	pUcscTrack.script      = "file:scripts/genomeplot/pUcscTrack.r"
	return pUcscTrack

@procfactory
def _pGenomePlot():
	"""
	@name:
		pGenomePlot
	@description:
		plot the genomic features
	@input:
		`trkfiles:files`: the list of track dumped files
		`region`:         the region, in format of `chr1:1-1000`
		`highlight`:      the highlight regions, informat of start1-end1; start2-end2; ...
	@output:
		`outfile:file`:   the figure
	@args:
		`genome`  : The genome
		`showIdeo`: Show ideogram track? Default: True
		`showAxis`: Show axis? Default: True
		`showGenes`: Show geneTrack? Default: True
		`params`:   The params
			- `genneral`:  General params for plotTracks
			- `geneTrack`: The params for geneTrack
	@requires:
		[r-Gviz](https://rdrr.io/bioc/Gviz)
	"""
	pGenomePlot                = Proc(desc = 'Plot genome elements.')
	pGenomePlot.input          = "trkfiles:files, region, highlight"
	pGenomePlot.output         = "outfile:file:genomeplot_{{i.region | lambda x: x.replace(':', '-')}}.png"
	pGenomePlot.args.genome    = params.genome.value
	pGenomePlot.args.ideoTrack = params.cytoband.value # a file from ucsc (http://hgdownload.soe.ucsc.edu/goldenPath/hg19/database/cytoBand.txt.gz); or True to download it in runtime; or False to disable it
	pGenomePlot.args.axisTrack = True
	pGenomePlot.args.geneTrack = params.refgene.value # or False to disable
	pGenomePlot.args.params    = Box({
		'title.width': .8
	})
	pGenomePlot.args.scheme  = Box(
		GenomeAxisTrack = Box({
			'fontsize': 8
		}),
		GeneRegionTrack = Box({
			'fontsize'         : 12,
			'fill'             : "salmon",
			'col'              : "NULL",
			'showId'           : True,
			'shape'            : ["smallArrow", "arrow"],
			'size'             : .5,
			'background.title' : "#333333",
			"arrowHeadWidth"   : 30,
			"arrowHeadMaxWidth": 40
		}),
		AnnotationTrack = Box({
			'fontsize'        : 12,
			'background.title': "#333333",
			'fontcolor.item'  : "#333333"
		})
	)
	pGenomePlot.args.devpars = Box(res = 300, height = 300, width = 2000)
	pGenomePlot.envs.rimport = rimport
	pGenomePlot.lang         = params.Rscript.value
	pGenomePlot.script       = "file:scripts/genomeplot/pGenomePlot.r"
	return pGenomePlot


